# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-08-24 16:39:48 alex>
#

"""
init for advanced functions
"""

from .sds import SDS
from .base import Base
from .class_params import ClassParams
from .space import Space
from .device import Device
from .devif import DeviceInterface
from .network import Network
from .ipaddress import IpAddress
from .dns import DNS
from .dns_zone import DNS_zone
from .dns_record import DNS_record

from .device_tools import list_devices



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-08-25 15:17:23 alex>
#
# only for python v3

"""
SOLIDserver base object
"""

# import logging

from SOLIDserverRest.Exception import SDSError


__all__ = ["Base"]


# just container class, no need for methods
# pylint: disable=R0903
class Base:
    """ standard class for all objects in SDS """
    # ---------------------------

    def __init__(self, sds=None, name=None):
        """init the base object:
        """

        # if true, modification on object are pushed to SDS
        self.in_sync = False

        self.sds = sds
        self.set_sds(sds)
        self.myid = -1
        self.name = name
        self.set_name(name)
        self.params = {}
        self.additional_params = {}
        self.additional_where_params = {}

    # -------------------------------------
    def clean_params(self):
        """ clean the object params """
        self.in_sync = False
        self.sds = None
        self.myid = -1
        self.name = None
        self.params = {}

    # ---------------------------
    def set_sds(self, sds=None):
        """set the sds connection for this object"""
        self.sds = sds

    # ---------------------------
    def set_name(self, name=None):
        """set the name for this object"""
        if name is None:
            self.name = None
            return

        if isinstance(name, str):
            self.name = name
        else:
            self.name = None
            raise SDSError("name format not valid")

    # ---------------------------
    def set_sync(self):
        """set the object modification sync with SDS"""
        self.in_sync = True

    # ---------------------------
    def set_async(self):
        """set the object modification async with SDS,
           calling update() required"""
        self.in_sync = False

    # ---------------------------
    def __str__(self):
        """return the string notation of the base object"""
        return " sync: {}".format(self.in_sync)

    # -------------------------------------
    def _get_id_by_name(self, query, key, name, key_id=None):
        """get the ID from its name, return None if non existant"""

        params = {
            "WHERE": "{}_name='{}'".format(key, name),
            "limit": 1,
            **self.additional_params
        }

        # logging.info(self.additional_where_params)
        for _key, _value in self.additional_where_params.items():
            if isinstance(_value, str):
                params['WHERE'] += " and {}='{}'".format(_key, _value)
            elif isinstance(_value, int):
                params['WHERE'] += " and {}={}".format(_key, _value)

        # pylint: disable=E1101
        if hasattr(self, 'space'):
            if self.space:
                params['WHERE'] += " and site_id={}".format(self.space.myid)

        # specific for networks
        if hasattr(self, 'subnet_addr'):
            if self.subnet_addr:
                params['WHERE'] += " and start_hostaddr="
                params['WHERE'] += "'{}'".format(self.subnet_addr)

        if hasattr(self, 'parent_network'):
            if self.parent_network:
                params['WHERE'] += " and parent_subnet_id="
                params['WHERE'] += "'{}'".format(self.parent_network.myid)
            elif 'parent_subnet_id' in self.params:
                if self.params['parent_subnet_id']:
                    _psid = self.params['parent_subnet_id']
                    params['WHERE'] += " and parent_subnet_id="
                    params['WHERE'] += "'{}'".format(_psid)

        # logging.info(query)
        # logging.info(params)

        try:
            rjson = self.sds.query(query,
                                   params=params)
        except SDSError as err_descr:
            msg = "cannot found object by name {}_name={}".format(key, name)
            msg += " / "+str(err_descr)
            raise SDSError(msg) from err_descr

        if rjson[0]['errno'] != '0':  # pragma: no cover
            raise SDSError("errno raised on get id by name")

        if key_id:
            return rjson[0]['{}_id'.format(key_id)]

        return rjson[0]['{}_id'.format(key)]

    # -------------------------------------
    def _get_id(self, query, key):
        """get the ID for the current object based
           on its current name
        """

        if self.myid >= 0:
            return self.myid

        if self.params['{}_id'.format(key)] is None:
            _id = self._get_id_by_name(query=query,
                                       key=key,
                                       name=self.name)

        self.myid = int(_id)

        return self.myid

    # -------------------------------------
    def set_param(self, param=None, value=None, exclude=None, name=None):
        """ set a specific param value """
        if param is None or not isinstance(param, str):
            return

        if value is None:
            return

        if param not in self.params:
            return

        b_do_set = True

        # exclude
        if exclude is not None:
            if param in exclude:
                b_do_set = False

        if b_do_set:
            self.params[param] = value
            if param == name:
                self.name = value

        if self.in_sync:
            self.update()

    # -------------------------------------
    def set_additional_params(self, **kwargs):
        """set any kind of additionnal parameter, may need to filter entries"""
        if self.additional_params == {}:
            self.additional_params = kwargs
        else:
            self.additional_params.update(kwargs)

    # -------------------------------------
    def clean_additional_params(self):
        """clean additionnal parameters on this object"""
        self.additional_params = {}

    # -------------------------------------
    def set_additional_where_params(self, **kwargs):
        """set any kind of additionnal parameter for where queries"""
        if self.additional_where_params == {}:
            self.additional_where_params = kwargs
        else:
            self.additional_where_params.update(kwargs)

    # -------------------------------------
    def clean_additional_where_params(self):
        """clean additionnal parameters on this object"""
        self.additional_where_params = {}

    # -------------------------------------
    def update(self):
        """ update template """

    # -------------------------------------
    def str_params(self, exclude=None):
        """ add params value to str"""

        return_val = " id={}".format(self.myid)

        sep = " "
        for key, value in self.params.items():
            if exclude is not None:
                if key in exclude:
                    continue

            if value == "":
                continue

            return_val += "{}{}={}".format(sep, key, value)
            sep = ", "

        return return_val



#
# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-08-24 16:00:39 alex>
#

"""
SOLIDserver DNS server management

"""

import ipaddress
import logging
import time

from SOLIDserverRest.Exception import (SDSInitError,
                                       SDSError,
                                       SDSDNSError,
                                       SDSDNSAlreadyExistingError,
                                       SDSDNSCredentialsError)

from .class_params import ClassParams


class DNS(ClassParams):
    """ class to manipulate the SOLIDserver DNS object """

    # -------------------------------------
    def __init__(self, sds=None,
                 name=None,
                 class_params=None):
        """init the DNS object:
        - sds: object SOLIDserver, could be set afterwards
        - name: dns server name
        """

        super(DNS, self).__init__(sds, name)

        self.ipv4_addr = None
        self.insync_wait_time = 0.5

        self.params = {
            'aws_keyid': None,
            'dns_allow_query': None,
            'dns_allow_query_cache': None,
            'dns_allow_recursion': None,
            'dns_allow_transfer': None,
            'dns_also_notify': None,
            'dns_class_name': '',
            'dns_comment': None,
            'dns_forward': '',
            'dns_forwarders': '',
            'dns_id': None,
            'dns_key_name': None,
            'dns_key_proto': None,
            'dns_key_value': None,
            'dns_name': None,
            'dns_notify': None,
            'dns_recursion': 'yes',
            'dns_role': None,
            'dns_state': None,
            'dns_synching': None,
            'dns_type': None,
            'dns_version': None,
            'dnsblast_enabled': None,
            'dnsblast_status': None,
            'dnsgslb_supported': None,
            'dnsguardian_supported': None,
            'dnssec_validation': None,
            'gss_enabled': None,
            'gss_keytab_id': None,
            'ip_addr': None,
            'ipmdns_https_login': None,
            'ipmdns_https_password': None,
            'ipmdns_is_package': None,
            'ipmdns_protocol': 'https',
            'ipmdns_type': None,
            'isolated': '0',
            'ldap_domain': None,
            'ldap_password': None,
            'ldap_user': None,
            'multistatus': None,
            'querylog_state': None,
            'total_vdns_members': None,
            'vdns_arch': None,
            'vdns_members_name': None,
            'vdns_parent_arch': None,
            'vdns_parent_id': None,
            'vdns_parent_name': None,
            # 'vdns_public_ns_list': None,
        }

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    def _get_dnsid_by_name(self, dnsname):
        """get the DNS server ID from its name,
           return None if non existant"""

        try:
            rjson = self.sds.query("dns_server_list",
                                   params={
                                       "WHERE": "dns_name='{}'".
                                                format(dnsname),
                                       **self.additional_params
                                   })
        except SDSError:
            return None

        if rjson[0]['errno'] != '0':
            raise SDSDNSError("dns_id errno raised")

        return rjson[0]['dns_id']

    # -------------------------------------
    def set_ipv4(self, ipv4=None):
        ''' set ipv4 address of the DNS server'''
        try:
            self.ipv4_addr = ipaddress.IPv4Address(ipv4)
        except ipaddress.AddressValueError:
            raise SDSDNSError(message='bad IPv4 address for DNS server')

    # -------------------------------------
    def set_type(self, newtype=None, vdns_arch=None):
        '''set type of the DNS server'''
        if newtype not in ['ipm', 'msdaemon', 'ans', 'aws', 'other', 'vdns']:
            raise SDSDNSError(message='bad DNS type')

        self.params['dns_type'] = newtype

        if newtype == 'vdns':
            if not vdns_arch:
                raise SDSDNSError(message='SMARTarchitecture needs type')
            if vdns_arch not in ['masterslave',
                                 'stealth',
                                 'multimaster',
                                 'single',
                                 'farm']:
                raise SDSDNSError(message='bad SMARRTarchitecture type')
            self.params['vdns_arch'] = vdns_arch

    # -------------------------------------
    def set_ipm_credentials(self, user, passwd):
        '''set admin credentials to talk to a ipmdns'''
        self.params['ipmdns_https_login'] = user
        self.params['ipmdns_https_password'] = passwd

    # -------------------------------------
    def _update_params(self, params):
        if self.params['dns_forward'] is not None:
            params['dns_forward'] = self.params['dns_forward']
            params['dns_forwarders'] = self.params['dns_forwarders']
        else:  # pragma: no cover
            params['dns_forward'] = ''
            params['dns_forwarders'] = ''

        if self.params['dns_recursion'] is None:   # pragma: no cover
            params['dns_recursion'] = 'yes'
        else:
            params['dns_recursion'] = self.params['dns_recursion']

    # -------------------------------------
    def create(self):
        """creates the DNS server"""
        if self.sds is None:
            raise SDSInitError(message="not connected")

        dns_id = self._get_dnsid_by_name(self.name)
        if dns_id is not None:
            raise SDSDNSAlreadyExistingError(message="already"
                                             + " existant dns server")

        if self.params['dns_type'] is None:
            raise SDSDNSError(message="DNS server"
                              + " with bad type, use set_type")

        params = {
            'dns_name': self.name,
            'dns_type': self.params['dns_type'],
            **self.additional_params
        }

        if self.params['dns_type'] != 'vdns':
            if self.ipv4_addr is None:
                raise SDSDNSError(message="DNS server"
                                  + " without IP address, use set_ipv4")
            params['hostaddr'] = self.ipv4_addr

            if self.params['dns_type'] == 'ipm':
                params['ipmdns_https_login'] = self.params[
                    'ipmdns_https_login']
                params['ipmdns_https_password'] = self.params[
                    'ipmdns_https_password']

        else:
            params['vdns_arch'] = self.params['vdns_arch']

        self._update_params(params)

        self.prepare_class_params('dns', params)

        try:
            rjson = self.sds.query("dns_server_create",
                                   params=params)
        except SDSError:   # pragma: no cover
            raise SDSDNSError(message="create DNS server")

        if 'errno' in rjson:   # pragma: no cover
            if rjson['errno'] == '14401':
                raise SDSDNSCredentialsError(message=rjson['errmsg'])

            logging.info(rjson)
            raise SDSDNSError(message="DNS creation error: "
                              + rjson['errmsg'])

        if 'ret_oid' not in rjson[0]:   # pragma: no cover
            raise SDSDNSError(message="DNS server creation error "
                              + rjson['errmsg'])

        self.params['dns_id'] = int(rjson[0]['ret_oid'])
        self.myid = self.params['dns_id']

        self.refresh()

    # -------------------------------------
    def delete(self):
        """deletes the DNS server in the SDS"""
        if self.sds is None:
            raise SDSDNSError(message="not connected")

        if self.myid is None or self.myid == -1:
            raise SDSDNSError(message="missing DNS id")

        self._wait_for_synch(self.myid)

        try:
            rjson = self.sds.query("dns_server_delete",
                                   params={
                                       'dns_id': self.myid,
                                       **self.additional_params
                                   })
            if 'errmsg' in rjson:  # pragma: no cover
                raise SDSDNSError(message="DNS server delete error, "
                                  + rjson['errmsg'])
        except SDSError:
            raise SDSDNSError(message="DNS server delete error")

        self.myid = -1

    # -------------------------------------
    def _wait_for_synch(self, dns_id=None):
        """waith for the DNS server to be in sync"""
        if dns_id is None:  # pragma: no cover
            return None

        _wait_delay = self.insync_wait_time

        for retry in range(10):
            rjson = self.sds.query("dns_server_info",
                                   params={
                                       "dns_id": dns_id,
                                   })

            if not rjson:   # pragma: no cover
                raise SDSDNSError(message="DNS in sync error")

            if rjson[0]['dns_synching'] == '0':
                if retry > 0:
                    self.insync_wait_time *= 1.2
                else:
                    self.insync_wait_time /= 1.1
                return rjson

            # logging.info('not in synch {}'.format(_wait_delay))
            time.sleep(_wait_delay)
            _wait_delay *= 2

        return None  # pragma: no cover

    # -------------------------------------
    def refresh(self):
        """refresh content of the DNS from the SDS"""
        if self.sds is None:
            raise SDSDNSError(message="not connected")

        if self.myid is None or self.myid == -1:
            dns_id = self._get_dnsid_by_name(self.name)
        else:
            dns_id = self.myid

        if dns_id is None:
            raise SDSDNSError(message="non existant DNS to refresh")

        rjson = self._wait_for_synch(dns_id)

        if not rjson:   # pragma: no cover
            raise SDSDNSError(message="DNS server refresh error, len of array")

        rjson = rjson[0]

        for label in [
                'aws_keyid',
                'dns_allow_query',
                'dns_allow_query_cache',
                'dns_allow_recursion',
                'dns_allow_transfer',
                'dns_also_notify',
                'dns_class_name',
                'dns_comment',
                'dns_forward',
                'dns_forwarders',
                'dns_hybrid',
                'dns_id',
                'dns_key_name',
                'dns_key_proto',
                'dns_key_value',
                'dns_name',
                'dns_notify',
                'dns_recursion',
                'dns_role',
                'dns_state',
                'dns_synching',
                'dns_type',
                'dns_version',
                'dnsblast_enabled',
                'dnsblast_status',
                'dnsgslb_supported',
                'dnsguardian_supported',
                'dnssec_validation',
                'gss_enabled',
                'gss_keytab_id',
                'ipmdns_is_package',
                'ipmdns_type',
                'isolated',
                'multistatus',
                'querylog_state',
                'vdns_arch',
                'vdns_members_name',
                'vdns_parent_name',
                'vdns_parent_id',
        ]:
            if label not in rjson:   # pragma: no cover
                raise SDSDNSError("parameter"
                                  + " {}".format(label)
                                  + " not found in DNS server")
            self.params[label] = rjson[label]

        if self.params['dns_type'] != 'vdns':
            self.ipv4_addr = ipaddress.IPv4Address(int(rjson['ip_addr'], 16))

        self.myid = int(self.params['dns_id'])

        if 'dns_class_parameters' in rjson:
            self.update_class_params(rjson['dns_class_parameters'])

    # -------------------------------------
    def update(self):
        """ update the DNS server """

        if self.sds is None:
            raise SDSDNSError(message="not connected")

        if self.myid is None or self.myid == -1:
            dns_id = self._get_dnsid_by_name(self.name)
        else:
            dns_id = self.myid

        if dns_id is None:
            raise SDSDNSError(message="non existant DNS to refresh")

        params = {
            'dns_id': dns_id,
            **self.additional_params
        }

        self._update_params(params)

        self.prepare_class_params('dns', params)

        # logging.info(params)

        rjson = self.sds.query("dns_server_update",
                               params=params)

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSDNSError(message="DNS server update error, "
                              + rjson['errmsg'])

        self.refresh()

    # -------------------------------------
    def set_forward(self, mode=None, targets=None):
        """ update the DNS server forwarders
            mode is: None, only or first
            targets is an array with ip addresses of targets as string
        """
        if mode is None:
            self.params['dns_forward'] = ''
            self.params['dns_forwarders'] = ''
        else:
            if mode in ['only', 'first']:
                self.params['dns_forward'] = mode
                self.params['dns_forwarders'] = ";".join(targets)
            else:
                raise SDSDNSError("bad forwarder type")  # pragma: no cover

    # -------------------------------------
    def set_recursion(self, mode=False):
        """ update the DNS server recursion mode
        """
        if mode:
            self.params['dns_recursion'] = 'yes'
        else:
            self.params['dns_recursion'] = 'no'

    # -------------------------------------
    def __str__(self):
        """return the string notation of the DNS server object"""
        return_val = "*DNS* name={}".format(self.name)

        if self.myid != -1:
            return_val += " id={}".format(self.myid)

        if self.ipv4_addr is not None:
            return_val += " ip={}".format(self.ipv4_addr)

        if self.params['dns_type'] == 'ipm':
            return_val += " ipm={}".format(str(self.params['ipmdns_type']))
        else:
            return_val += " type={}".format(str(self.params['dns_type']))

        if self.params['dns_recursion'] is not None:
            return_val += " recursion={}".format(self.params['dns_recursion'])

        if self.params['dns_forward'] in ['first', 'only']:
            return_val += " forward"
            return_val += "[{}]={}".format(self.params['dns_forward'],
                                           self.params['dns_forwarders'])

        return_val += " notify={}".format(str(self.params['dns_notify']))

        if self.params['dns_class_name'] != '':    # pragma: no cover
            return_val += " class="
            return_val += "{}".format(self.params['dns_class_name'])

        return_val += str(super(DNS, self).__str__())

        return return_val



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2020-05-14 21:44:19 alex>
#

"""
SOLIDserver device manager host interface

"""

import logging
import ipaddress
import time
import re

from SOLIDserverRest.Exception import SDSError
from SOLIDserverRest.Exception import SDSDeviceIfError
from SOLIDserverRest.Exception import SDSDeviceIfNotFoundError

from .class_params import ClassParams
from .device import Device
from .space import Space


# pylint: disable=R0902
class DeviceInterface(ClassParams):
    """ class to manipulate the SOLIDserver device interface """

    # -------------------------------------
    def __init__(self, sds=None, name=None, device=None, class_params=None):
        """init a device interface object:
        - sds: object SOLIDserver, could be set afterwards
        - name: name of the interface
        - device: reference to a Device object
        """

        if name is None:
            raise SDSDeviceIfError("no name provided at device interface init")

        if device is None or not isinstance(device, Device):
            raise SDSDeviceIfError("no device provided")

        if device.myid == -1:
            raise SDSDeviceIfError("device should be linked in SDS")

        super(DeviceInterface, self).__init__()

        # params mapping the object in SDS
        self.clean_params()

        self.set_sds(sds)
        self.set_name(name)
        self.set_device(device)

        self.ipv4 = None
        self.ipv6 = None
        self.mac = None

        self.space = None

        self.type = "interface"

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    def clean_params(self):
        """ clean the object params """
        super(DeviceInterface, self).clean_params()

        self.params = {
            'hostiface_id': None,
            'hostiface_name': None,
            'hostdev_id': None,
            'modify_time': None,
        }

        self.device = None
        self.ipv4 = None
        self.ipv6 = None
        self.mac = None
        self.space = None
        self.type = "interface"

    # -------------------------------------
    def set_param(self, param=None, value=None, exclude=None, name=None):
        super(DeviceInterface, self).set_param(param,
                                               value,
                                               exclude=['hostiface_id'],
                                               name='hostiface_name')

    # -------------------------------------
    def create(self):
        """ create the device interface in SDS """

        if self.sds is None:
            raise SDSDeviceIfError(message="not connected")

        params = {
            'hostiface_name': self.name,
            'hostdev_id': self.device.myid,
            'hostiface_type': self.type,
            **self.additional_params
        }

        if self.ipv4 is not None:
            params['hostiface_addr'] = self.ipv4

        if self.mac is not None:
            params['hostiface_mac'] = self.mac

        if (self.space is not None
                and isinstance(self.space, Space)
                and self.space.myid != -1):
            params['site_id'] = self.space.myid
        else:
            logging.warning("no space set on interface")

        self.prepare_class_params('hostiface', params)

        # logging.info(params)

        rjson = self.sds.query("host_iface_create",
                               params=params)

        if 'errmsg' in rjson:
            raise SDSDeviceIfError(message="devif creation error, "
                                   + rjson['errmsg'])

        self.params['hostiface_id'] = int(rjson[0]['ret_oid'])
        self.myid = int(self.params['hostiface_id'])

        self.refresh()

    # -------------------------------------
    def update(self):
        """ update the device interface in SDS """

        if self.sds is None:
            raise SDSDeviceIfError(message="not connected")

        params = {
            'hostiface_id': self._get_id(query="host_iface_list",
                                         key="hostiface"),
            'hostiface_name': self.name,
            'modify_time': int(time.time()),
            **self.additional_params
        }

        if self.ipv4 is not None:
            params['hostiface_addr'] = self.ipv4

        self.prepare_class_params('hostiface', params)

        rjson = self.sds.query("host_iface_update",
                               params=params)

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSDeviceIfError(message="interface update error, "
                                   + rjson['errmsg'])

        self.refresh()

    # -------------------------------------
    def delete(self):
        """deletes the device interface in the SDS"""
        if self.sds is None:
            raise SDSDeviceIfError(message="not connected")

        if self.myid == -1:
            raise SDSDeviceIfNotFoundError("on delete")

        params = {
            'hostiface_id': self.myid,
            **self.additional_params
        }

        rjson = self.sds.query("host_iface_delete",
                               params=params)

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSDeviceIfNotFoundError("on delete "+rjson['errmsg'])

        self.clean_params()

    # -------------------------------------
    def refresh(self):
        """refresh content of the device interface from the SDS"""

        if self.sds is None:
            raise SDSDeviceIfError(message="not connected")

        if self.myid == -1:
            params = {
                **self.additional_params
            }

            params['WHERE'] = "hostiface_name='{}'".format(self.name)
            params['WHERE'] += " and hostdev_name"
            params['WHERE'] += "='{}'".format(self.device.name)

            try:
                rjson = self.sds.query("host_iface_list",
                                       params=params)
            except SDSError as err_descr:  # pragma: no cover
                msg = "cannot found object by name"
                msg += " / "+str(err_descr)
                raise SDSError(msg)

            if rjson[0]['errno'] != '0':  # pragma: no cover
                raise SDSError("errno raised on get id by name")

            if_id = rjson[0]['hostiface_id']
        else:
            if_id = self.myid

        params = {
            "hostiface_id": if_id,
            **self.additional_params
        }

        try:
            rjson = self.sds.query("host_iface_info",
                                   params=params)
        except SDSError as err_descr:   # pragma: no cover
            msg = "cannot get device interface info on id={}".format(if_id)
            msg += " / "+str(err_descr)
            raise SDSDeviceIfError(msg)

        rjson = rjson[0]
        # logging.info(rjson)

        for label in ['hostiface_id',
                      'hostiface_name',
                      'hostdev_id',
                      'iplnetdev_id',
                      'hostiface_type',
                      'hostiface_class_name',
                      'pear_iface_id',
                      'pear_ipl_iface_id',
                      'iplport_id',
                      'custom_db_data_id',
                      'hostiface_mac',
                      'add_time',
                      'modify_time',
                      'vendor_key',
                      'vendor_mac',
                      'hostiface_ip_formated',
                      'row_enabled',
                      'hostiface_site_id',
                      'hostiface_manual_link',
                      'hostiface_auto_link',
                      'nb_ip'
                      ]:
            if label not in rjson:  # pragma: no cover
                msg = "parameter {} not found in device".format(label)
                raise SDSDeviceIfError(msg)
            self.params[label] = rjson[label]

        self.myid = int(self.params['hostiface_id'])
        self.name = self.params['hostiface_name']

        if (int(self.params['hostdev_id'])
                != self.device.myid):  # pragma: no cover
            logging.error('interface has changed device - TODO')
            logging.error('%s / %s', self.device.myid,
                          self.params['hostdev_id'])

        if 'hostiface_class_parameters' in rjson:
            self.update_class_params(rjson['hostiface_class_parameters'])

        if 'hostiface_ip_formated' in rjson:
            # regexpip = re.compile('
            ip_raw = re.search(r'^(((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.)'
                               '{3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))'
                               ' .*$',
                               rjson['hostiface_ip_formated'])

            if ip_raw:
                self.ipv4 = ip_raw.group(1)

        self.device.link_interface(self)

    # ---------------------------
    def set_device(self, dev):
        """set the device"""
        self.device = dev

    # ---------------------------
    def set_ipv4(self, addr):
        """set the ipv4 address on the interface"""
        self.ipv4 = None
        try:
            self.ipv4 = str(ipaddress.IPv4Address(addr))
        except ValueError:
            raise SDSDeviceIfError('bad ipv4 format')

    # ---------------------------
    def set_mac(self, mac):
        """set the mac address on the interface"""
        self.mac = None
        self.mac = str(mac)

    # ---------------------------
    def set_space(self, space):
        """set the space on the interface"""
        self.space = None
        if not isinstance(space, Space):
            raise SDSDeviceIfError("space incorect")

        if space.myid == -1:
            raise SDSDeviceIfError("space not synchronized")

        self.space = space

    # -------------------------------------
    def __str__(self):  # pragma: no cover
        """return the string notation of the device interface object"""

        return_val = "*dev if* name={}".format(self.name)

        if self.ipv4 is not None:
            return_val += " ipv4={}".format(self.ipv4)

        if self.ipv6 is not None:
            return_val += " ipv6={}".format(self.ipv6)

        return_val += self.str_params(exclude=['hostiface_id',
                                               'hostiface_name'])

        return_val += str(super(DeviceInterface, self).__str__())

        return return_val



# -*- Mode: Python; python-indent-offset: 4 -*-
# -*-coding:Utf-8 -*
#
# Time-stamp: <2020-07-26 16:25:25 alex>
#
# disable naming convention issue
# pylint: disable=C0103
##########################################################
# Request example:
# http://<SOLIDserver-IP>/rest/<service>?<param> [param=URLencode(value)]
###########################################################

"""
Efficient IP low level SOLIDServer API binding

TODO: add tests for proxy
"""

import sys
import base64
import re
import logging
import urllib
import requests


# pylint: disable=F0401, E1101
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from OpenSSL import crypto  # not working on windows pylint: disable=E0401

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
# pylint: enable=F0401, E1101


if sys.version_info[0] == 2:   # pragma: no cover
    # pylint: disable=F0401
    from mapper import SERVICE_MAPPER, METHOD_MAPPER
    from Exception import SDSInitError, SDSError
    from Exception import SDSServiceError, SDSRequestError
    # pylint: enable=F0401
else:
    from .mapper import SERVICE_MAPPER, METHOD_MAPPER
    from .Exception import SDSInitError, SDSError
    from .Exception import SDSServiceError, SDSRequestError

__all__ = ["SOLIDserverRest"]

##########################################################################


# effectively few variables in this class, just disabling the warning
# pylint: disable=R0902
class SOLIDserverRest:
    """ main SDS class """
    CNX_NATIVE = 1
    CNX_APIKEY = 2
    CNX_BASIC = 3

    # -------------------------------------
    def __init__(self, host, debug=False):
        """ initialize connection with SDS host,
            this function is not active,
            just set host and parameters
        """

        self.auth = None
        self.cnx_type = None
        self.debug = debug
        self.headers = None
        self.host = host
        self.last_url = ''
        self.password = None
        self.resp = None
        self.user = None
        self.session = None
        self.default_method = 'rest'
        self.prefix_url = 'https://{}'.format(host)
        self.python_version = 0
        self.fct_url_encode = None
        self.fct_b64_encode = None
        self.ssl_verify = True
        self.proxies = None

        # set specific features for python v2 (<=2020, not supported after)
        if sys.version_info[0] == 2:   # pragma: no cover
            # pylint: disable=E1101
            self.python_version = 2
            self.fct_url_encode = urllib.urlencode
            self.fct_b64_encode = base64.standard_b64encode
            # pylint: enable=E1101
        else:
            self.python_version = 3
            self.fct_url_encode = urllib.parse.urlencode
            self.fct_b64_encode = base64.b64encode

        self.last_url = ''
        self.resp = None

        self.session = requests.Session()
        # self.session.verify = "cert.pem"

    # -------------------------------------
    def __del__(self):
        self.clean()

    # -------------------------------------
    def use_native_sds(self, user, password):
        """ propose to use a native EfficientIP SDS connection with Username
        and password encoded in the headers of each requests
        """
        logging.debug("useNativeSDS %s %s", user, password)

        # check if SDS connection is established
        if self.host is None:
            raise SDSInitError()

        self.user = user
        self.password = password
        self.cnx_type = self.CNX_NATIVE

        # Encryption management in function of Python version
        self.headers = {
            'X-IPM-Username': self.fct_b64_encode(user.encode()),
            'X-IPM-Password': self.fct_b64_encode(password.encode()),
            'content-type': 'application/json'
        }

    # -------------------------------------
    def use_basicauth_sds(self, user, password):
        """ propose to use the basic auth implementation on the SDS
        """
        logging.debug("useBasicAuthSDS %s %s", user, password)

        # check if SDS connection is established
        if self.host is None:
            raise SDSInitError()

        self.user = user
        self.password = password

        self.cnx_type = self.CNX_BASIC
        self.session.auth = requests.auth.HTTPBasicAuth(user, password)

        self.headers = {
            'content-type': 'application/json'
        }

    # -------------------------------------
    def set_certificate_file(self, file_path):
        """set the certificate that will be used to authenticate the server"""
        try:
            file_content = open(file_path, 'r').read()
            crypto.load_certificate(crypto.FILETYPE_PEM,
                                    file_content)
        except IOError:
            logging.warning("cannot load CA file")
            raise SDSInitError("cannot load CA file {}".format(file_path))
        except crypto.Error as error:
            logging.warning(error)
            raise SDSInitError("invalid CA file {}".format(file_path))

        self.session.verify = file_path
        self.ssl_verify = True

    # -------------------------------------
    def set_ssl_verify(self, value):
        """allows to enable or disable the certificate validation"""
        if isinstance(value, bool):
            self.ssl_verify = value
        else:
            logging.warning("bad type when calling set_ssl_verify")
            raise SDSError("requested bool on set_ssl_verify")

    # -------------------------------------
    def set_proxy(self, proxy):   # pragma: no cover
        """allows to enable or disable use of a SOCKS proxy"""
        proxy = 'socks5h://{}'.format(proxy)
        self.proxies = {'http': proxy, 'https': proxy}

    # -------------------------------------
    def _query_method(self, service, option, params):
        if params != '':
            params = "?"+self.fct_url_encode(params)

        # choose method
        method = None
        if option:
            method = 'OPTIONS'
            params = ''
        else:
            for verb in METHOD_MAPPER:
                _q = ".*_{}$".format(verb)
                if re.match(_q, service) is not None:
                    method = METHOD_MAPPER[verb]
                    break

        if method is not None:
            # flag_add management
            if method == 'POST':
                params = "{}{}".format(params, '&add_flag=new_only')
            elif method == 'PUT':
                params = "{}{}".format(params, '&add_flag=edit_only')

        return (params, method)

    # -------------------------------------
    def query(self, service,
              params='',
              timeout=2,
              option=False):
        """ send request to the API endpoint, returns request result """

        (params, method) = self._query_method(service, option, params)

        if method is None:
            msg = "no method available for request {}".format(service)
            logging.warning("no method available for request %s", service)
            raise SDSServiceError(service,
                                  message=msg)

        logging.debug("method %s selected for service %s", method, service)

        # choose service
        svc_mapped = SERVICE_MAPPER.get(service)
        if svc_mapped is None:
            logging.warning("unknown service %s", service)
            raise SDSServiceError(service)

        self.last_url = "{}{}".format(svc_mapped, params).strip()

        if re.match('.*_find_free$', service) is not None:
            url = "{}/{}/{}".format(self.prefix_url,
                                    'rpc',
                                    self.last_url)
        else:
            url = "{}/{}/{}".format(self.prefix_url,
                                    self.default_method,
                                    self.last_url)

        try:
            logging.debug("m=%s u=%s h=%s v=%s a=%s",
                          method,
                          url,
                          self.headers,
                          self.ssl_verify,
                          self.auth)

            return self.session.request(
                method,
                url,
                headers=self.headers,
                proxies=self.proxies,
                verify=self.ssl_verify,
                timeout=timeout,
                auth=self.auth)
        except requests.exceptions.SSLError:
            raise SDSRequestError(method,
                                  url,
                                  self.headers,
                                  message="SSL certificate error")
        except BaseException as error:   # pragma: no cover
            raise SDSRequestError(method, url, self.headers, message=error)

    # -------------------------------------
    def get_headers(self):
        """ returns the headers attached to this connection """
        return self.headers

    # -------------------------------------
    def get_proxies(self):   # pragma: no cover
        """ returns the proxies attached to this connection """
        return self.proxies

    # -------------------------------------
    def get_status(self):
        """ returns status of the SDS connection """
        _r = {
            'host': self.host,
            'python_version': self.python_version
        }
        return _r

    # -------------------------------------
    def clean(self):
        """ clean all status of the SDS connection """
        self.auth = None
        self.cnx_type = None
        self.debug = None
        self.headers = None
        self.host = None
        self.last_url = ''
        self.password = None
        self.default_method = 'rest'
        self.prefix_url = None
        self.python_version = None
        self.resp = None
        self.user = None
        self.session = None
        self.ssl_verify = True
        self.proxies = None

    # -------------------------------------
    def __str__(self):   # pragma: no cover
        _s = "SOLIDserverRest: API={}, user={}"
        return(_s.format(self.prefix_url,
                         self.user))

    # deprecated method to be suppressed

    def use_native_ssd(self, user, password):   # pragma: no cover
        """deprecated version of use_native_sds"""
        logging.critical("deprecated method use_native_ssd")
        self.use_native_sds(user, password)

    def use_basicauth_ssd(self, user, password):   # pragma: no cover
        """deprecated version of use_basicauth_ssd"""
        logging.critical("deprecated method use_basicauth_ssd")
        self.use_basicauth_sds(user, password)



#
# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-08-25 16:09:17 alex>
#

"""
SOLIDserver DNS record management

"""

# import logging
import time
import ipaddress
# import pprint

from SOLIDserverRest.Exception import (SDSInitError,
                                       SDSError,
                                       SDSDNSError)

from .class_params import ClassParams
from .dns_zone import DNS_zone
from .sds import SDS


class DNS_record(ClassParams):  # pylint: disable=C0103
    """ class to manipulate a DNS record object, from a zone """
    # -------------------------------------

    def __init__(self, sds=None, name=None, rr_type=None,
                 class_params=None):
        """init the record object"""

        if sds and not isinstance(sds, SDS):
            raise SDSInitError(message="sds param is not of type SDS")

        super(DNS_record, self).__init__(sds, name)

        self.zone = None
        self.ttl = 3600
        self.values = {}
        self.rr_type = None

        if rr_type:
            self.rr_type = str(rr_type)

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    def set_zone(self, zone):
        """ link the record to a zone on the dns server """
        if not isinstance(zone, DNS_zone):
            raise SDSDNSError(message="record needs to be attached to a zone")

        self.zone = zone

    # -------------------------------------
    def set_type(self, rr_type, **kvargs):
        """ set the type for the record
            args depending on record type:
             * A/AAAA: ip
             * MX: priority, target
             * TXT: txt
             * NS: target
             * CNAME: target
        """

        self.rr_type = str(rr_type)

        # logging.info(kvargs)

        if len(kvargs) == 0:
            return

        if self.rr_type == 'A':
            self.set_values([kvargs['ip']])
        elif self.rr_type == 'AAAA':
            self.set_values([kvargs['ip']])
        elif self.rr_type == 'MX':
            self.set_values([kvargs['priority'],
                             kvargs['target']])
        elif self.rr_type == 'TXT':
            self.set_values([kvargs['txt']])
        elif self.rr_type == 'NS':
            self.set_values([kvargs['target']])
        elif self.rr_type == 'CNAME':
            self.set_values([kvargs['target']])

    # -------------------------------------

    def set_values(self, avalues):
        """ set the values depending on the record type """
        if not self.rr_type:
            raise SDSDNSError(message='need to set type of'
                              ' record before values')

        if self.rr_type == 'A':
            # check v1 is an ip address
            try:
                ipaddress.IPv4Address(avalues[0])
            except ipaddress.AddressValueError:
                raise SDSInitError(message="record A requires an IP address")

            self.values = {
                '1': avalues[0]
            }

        elif self.rr_type == 'AAAA':
            # check v1 is an ip address
            try:
                ipaddress.IPv6Address(avalues[0])
            except ipaddress.AddressValueError:
                raise SDSInitError(message="record A requires an IP address")

            self.values = {
                '1': avalues[0]
            }

        elif self.rr_type == 'TXT':
            self.values = {
                '1': str(avalues[0])
            }

        elif self.rr_type == 'NS':
            self.values = {
                '1': str(avalues[0])
            }

        elif self.rr_type == 'CNAME':
            self.values = {
                '1': str(avalues[0])
            }

        elif self.rr_type == 'MX':
            self.values = {
                '1': int(avalues[0]),
                '2': str(avalues[1])
            }

        else:
            raise SDSDNSError(message="unknown type"
                              " of record {}".format(self.rr_type))

    # -------------------------------------
    def set_ttl(self, ttl):
        """ set the ttl for this record """

        self.ttl = int(ttl)
        if self.ttl < 5:
            self.ttl = 5

    # -------------------------------------
    def create(self, sync=True):
        """creates the DNS record in the zone"""
        if self.sds is None:
            raise SDSInitError(message="not connected")

        if self.zone is None:
            raise SDSDNSError(message="zone not attached")

        if not self.rr_type:
            raise SDSDNSError(message="record type not set")

        if '1' not in self.values:
            raise SDSDNSError(
                message="no values set for record {}".format(self.rr_type))

        params = {
            'rr_name': self.name,
            'rr_type': self.rr_type,
            'dns_id': self.zone.dns_server.myid,
            'dnszone_id': self.zone.myid,
            'rr_ttl': str(self.ttl),
            'value1': self.values['1'],
            **self.additional_params
        }

        for _v in ['2', '3', '4', '5', '6', '7']:
            if _v in self.values:
                params['value{}'.format(_v)] = self.values[_v]

        self.prepare_class_params('rr', params)

        try:
            rjson = self.sds.query("dns_rr_create",
                                   params=params)
        except SDSError:   # pragma: no cover
            raise SDSDNSError(message="create DNS record")

        # logging.info(rjson)
        if 'errno' in rjson and int(rjson['errno']) > 0:
            raise SDSDNSError(message="record:"
                              " {}".format(rjson['errmsg']))

        rjson = rjson[0]
        if 'ret_oid' in rjson:
            self.myid = int(rjson['ret_oid'])

        if sync:
            self.refresh()

    # -------------------------------------
    def _wait_for_synch(self, delete=False):
        """wait for the DNS record to be in sync"""
        if self.myid is None or self.myid == -1:
            raise SDSDNSError(message="missing DNS record id")

        _wait_delay = 0.1

        for _ in range(10):
            try:
                rjson = self.sds.query("dns_rr_info",
                                       params={
                                           "rr_id": self.myid,
                                       })
            except SDSError:
                return None

            if not rjson:   # pragma: no cover
                raise SDSDNSError(message="DNS record sync error")

            if not delete:
                # we wait for the zone to be pushed to the server
                if rjson[0]['delayed_create_time'] == '0':
                    return rjson
            else:
                # we wait for the zone to be deleted from the server
                if rjson[0]['delayed_delete_time'] == '0':
                    return None

            # logging.info('not yet in synch %s %f', self.name, _wait_delay)
            time.sleep(_wait_delay)
            _wait_delay *= 2

        raise SDSDNSError(message="DNS record sync takes too long")

    # -------------------------------------
    def refresh(self):
        """refresh content of the DNS record from the SDS"""
        if self.sds is None:
            raise SDSDNSError(message="not connected")

        if self.zone is None:
            raise SDSDNSError(message="zone not attached")

        if self.myid is None or self.myid == -1:
            self.set_additional_where_params(rr_type=self.rr_type)
            self.set_additional_where_params(dns_id=self.zone.dns_server.myid)
            self.set_additional_where_params(dnszone_name=self.zone.name)

            rr_id = self._get_id_by_name('dns_rr_list',
                                         'rr_full',
                                         self.name,
                                         key_id='rr')

            self.clean_additional_where_params()
        else:
            rr_id = self.myid

        if rr_id is None:
            raise SDSDNSError(message="non existant DNS record to refresh")

        self.myid = rr_id
        rjson = self._wait_for_synch()

        if not rjson:   # pragma: no cover
            raise SDSDNSError(message="DNS record refresh error, len of array")

        rjson = rjson[0]

        for label in [
                'dns_class_name',
                'dns_cloud',
                'dns_comment',
                'dns_type',
                'dns_version',
                'rr_class_name',
                'rr_full_name_utf',
                'rr_glue',
                'ttl',
                'value1',
                'value2',
                'value3',
                'value4',
                'value5',
                'value6',
                'value7'
        ]:
            if label not in rjson:   # pragma: no cover
                raise SDSDNSError("parameter"
                                  + " {}".format(label)
                                  + " not found in DNS zone")
            self.params[label] = rjson[label]

        if 'rr_class_parameters' in rjson:
            self.update_class_params(rjson['rr_class_parameters'])

        self.ttl = int(rjson['ttl'])
        self.set_values([rjson['value1'],
                         rjson['value2'],
                         rjson['value3'],
                         rjson['value4'],
                         rjson['value5'],
                         rjson['value6'],
                         rjson['value7']])

    # -------------------------------------
    def delete(self, sync=True):
        """deletes the DNS record from the zone"""
        if self.sds is None:
            raise SDSDNSError(message="not connected")

        if self.myid is None or self.myid == -1:
            raise SDSDNSError(message="missing DNS RR id")

        try:
            rjson = self.sds.query("dns_rr_delete",
                                   params={
                                       'rr_id': self.myid,
                                       **self.additional_params
                                   })
            if 'errmsg' in rjson:  # pragma: no cover
                raise SDSDNSError(message="DNS record delete, "
                                  + rjson['errmsg'])
        except SDSError:
            raise SDSDNSError(message="DNS record delete error")

        if sync:
            time.sleep(0.1)
            self._wait_for_synch(delete=True)

        self.myid = -1

    # -------------------------------------
    def __str__(self):
        """return the string notation of the DNS record object"""
        return_val = "*RR* name={}".format(self.name)

        if self.myid and self.myid != -1:
            return_val += " [#{}]".format(self.myid)

        if self.zone:
            return_val += " server={}".format(self.zone.dns_server.name)
            return_val += " zone={}".format(self.zone.name)

        if self.rr_type:
            return_val += " {}".format(self.rr_type)
            if self.rr_type == 'A':
                if '1' in self.values:
                    return_val += "={}".format(self.values['1'])

            return_val += " ttl={}".format(self.ttl)

            if 'rr_glue' in self.params:
                return_val += " glue={}".format(self.params['rr_glue'])

        return_val += str(super(DNS_record, self).__str__())

        return return_val



# -*- Mode: Python; python-indent-offset: 4 -*-
# -*-coding:Utf-8 -*
#
# Time-stamp: <2019-09-22 15:39:20 alex>
#
# disable naming convention issue
# pylint: disable=C0103
##########################################################

"""
init of the SOLIDserver module
"""

import sys

if sys.version_info[0] == 2:   # pragma: no cover
    # pylint: disable=F0401, W0406
    from SOLIDserverRest import *
    from mapper import *
    from Exception import *
    # pylint: enable=F0401, W0406
else:
    from .mapper import *
    from .SOLIDserverRest import *
    from .Exception import *



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2020-05-14 21:52:24 alex>
#

"""
SOLIDserver device manager

"""

import ipaddress
import re

# import logging
# import pprint
# import time

from SOLIDserverRest.Exception import SDSError, SDSInitError
from SOLIDserverRest.Exception import SDSDeviceError
from SOLIDserverRest.Exception import SDSEmptyError

# from .class_params import ClassParams
# from . import Device
# from . import DeviceInterface


# ------------------------------------------
def apply_filter_metadata(filt, dev_raw):
    """apply metadata value filter"""

    k = "tag_hostdev_{}".format(filt['name'])

    return dev_raw[k] == str(filt['val'])


# ------------------------------------------
def apply_filter_of_class(filt, dev_raw):
    """apply class parameter filter"""
    return dev_raw['hostdev_class_name'] == filt['val']


# ------------------------------------------
def apply_filter_in_subnet(filt, dev_raw):
    """apply ip address filter"""
    result = True

    if '_ipnet' not in filt:
        ipnet = ipaddress.ip_network(filt['val'])
        filt['_ipnet'] = ipnet
    else:
        ipnet = filt['_ipnet']

    if '_ipadd' not in dev_raw:
        dev_raw['_ipadd'] = []
        ip_raw = re.findall(r'(((25[0-5]|2[0-4][0-9]|[01]?'
                            r'[0-9][0-9]?)\.){3}(25[0-5]|'
                            r'2[0-4][0-9]|[01]?[0-9][0-9]?))',
                            dev_raw['hostdev_ip_formated'])
        for host_if_ip in ip_raw:
            if host_if_ip[0] not in dev_raw['_ipadd']:
                dev_raw['_ipadd'].append(host_if_ip[0])

    ip_raw = dev_raw['_ipadd']

    for host_if_ip in ip_raw:
        ipadd = ipaddress.ip_address(host_if_ip)

        if ipadd in ipnet:  # pylint: disable=R1723
            result = True
            break
        else:
            result = False

    return result


# ------------------------------------------
def apply_filter_space(filt, dev_raw):
    """apply space on ip address filter"""
    space_re = r'&([^,;]+)[,;]'

    spaces = re.findall(space_re, dev_raw['hostdev_ip_addr'])
    for space in spaces:
        if space == filt['val']:
            return True

    return False


# ------------------------------------------
def apply_list_filters(filters, dev_raw):
    """apply the list filters to the device and return True if all match"""
    result = True

    for filt in filters:
        if filt['type'] == "in_subnet":
            result = apply_filter_in_subnet(filt, dev_raw)
            if not result:
                return False

        if filt['type'] == "of_class":
            result = apply_filter_of_class(filt, dev_raw)
            if not result:  # pragma: no cover
                return False

        if filt['type'] == "metadata":
            result = apply_filter_metadata(filt, dev_raw)
            if not result:  # pragma: no cover
                return False

        if filt['type'] == "space":
            result = apply_filter_space(filt, dev_raw)
            if not result:
                return False

    return result


# ------------------------------------------
def _handle_filters(filters, _metadata_tags, _params):
    if filters is None:  # pragma: no cover
        filters = []

    # apply filters in the original query when possible
    for filt in filters:
        if filt['type'] == "of_class":
            _class_filter = "hostdev_class_name='{}'".format(filt['val'])
            if 'WHERE' in _params:
                _params['WHERE'] += " and " + _class_filter
            else:
                _params['WHERE'] = _class_filter

        if filt['type'] == "metadata" and filt['val'] not in _metadata_tags:
            _metadata_filter = "hostdev.{}".format(filt['name'])
            if 'TAGS' in _params:
                _params['TAGS'] += "&"+_metadata_filter
            else:
                _params['TAGS'] = _metadata_filter

            _metadata_filter = "tag_hostdev_{}='{}'".format(filt['name'],
                                                            str(filt['val']))
            if 'WHERE' in _params:
                _params['WHERE'] += " and " + _metadata_filter
            else:
                _params['WHERE'] = _metadata_filter

            _metadata_tags.append(filt['name'])


# ------------------------------------------
def _handle_metadata(_metadata_tags, _metadatas, _params):
    # expose metadata using TAGS
    if _metadatas:
        for _md in _metadatas:
            if _md not in _metadata_tags:
                _metadata_filter = "hostdev.{}".format(_md)
                if 'TAGS' in _params:
                    _params['TAGS'] += "&"+_metadata_filter
                else:
                    _params['TAGS'] = _metadata_filter


# ------------------------------------------
def list_devices_page(sds=None,
                      filters=None,
                      metadatas=None,
                      offset=0, page=25):
    """list devices from the space"""
    metadata_tags = []

    if sds is None:  # pragma: no cover
        raise SDSInitError("list device requires connection")

    params = {
        'limit': page,
        'offset': offset
    }

    _handle_filters(filters, metadata_tags, params)
    _handle_metadata(metadata_tags, metadatas, params)

    try:
        rjson = sds.query("host_device_list",
                          params=params)
    except SDSEmptyError:
        return None

    if 'errmsg' in rjson:  # pragma: no cover
        raise SDSError(message="device list, "
                       + rjson['errmsg'])

    adevs = []
    for dev in rjson:
        res = None

        # check the filters
        if apply_list_filters(filters, dev):
            res = {
                'name': dev['hostdev_name'],
            }
            if '_ipadd' in dev:
                res['extracted_ips'] = dev['_ipadd']
            else:  # pragma: no cover
                res['hostdev_ip_formated'] = dev['hostdev_ip_formated']

        # expose metadata if any
        if res and metadatas:
            for _md in metadatas:
                tag = "tag_hostdev_{}".format(_md)
                if tag in dev and dev[tag] != '':
                    res[_md] = dev[tag]

        if res:
            adevs.append(res)

    return adevs


# ------------------------------------------
def list_devices(sds=None,
                 filters=None,
                 metadatas=None,
                 limit=10):
    """ limit: max number of device to return, 0 for all
        filters: returned device should match all filters
            {'type':'in_subnet', 'val': '10.149.0.0/16'},
            {'type':'of_class', 'val': 'AWS-EC2'},
            {'type':'metadata', 'name': 'cores', 'val': '1'},
            {'type':'space', 'val': 'ex-space-01'},
    """

    offset = 0
    collected = 0
    page = 500
    alldevs = []

    if sds is None:
        raise SDSInitError(message="no sds object in list_devices")

    if filters is None:  # pragma: no cover
        filters = []

    # if no filter on IP address, add one
    ip_filt = False
    for filt in filters:
        if 'type' not in filt:
            raise SDSDeviceError(message="list filter without type")
        if filt['type'] == "in_subnet":
            ip_filt = True

    if not ip_filt:
        filters.append({'type': 'in_subnet', 'val': '0.0.0.0/0'})

    while limit == 0 or collected < limit:
        adevs = list_devices_page(sds,
                                  filters,
                                  metadatas,
                                  offset=offset,
                                  page=page)

        if adevs is None:
            break

        collected += len(adevs)
        alldevs += adevs

        offset += page

    if limit and len(alldevs) > limit:
        alldevs = alldevs[:limit]

    return alldevs



#!/usr/bin/python
# -*-coding:Utf-8 -*
##############################################

"""
set of values used to map calls from names to API endpoints
and method to use in REST calls
"""
__all__ = ["SERVICE_MAPPER", "METHOD_MAPPER"]

SERVICE_MAPPER = {
    'ip_site_create': 'ip_site_add',
    'ip_site_update': 'ip_site_add',
    'ip_site_count': 'ip_site_count',
    'ip_site_list': 'ip_site_list',
    'ip_site_info': 'ip_site_info',
    'ip_site_delete': 'ip_site_delete',
}

SERVICE_MAPPER.update({
    'ip_subnet_create': 'ip_subnet_add',
    'ip_subnet_update': 'ip_subnet_add',
    'ip_subnet_count': 'ip_block_subnet_count',
    'ip_subnet_list': 'ip_block_subnet_list',
    'ip_subnet_info': 'ip_block_subnet_info',
    'ip_subnet_delete': 'ip_subnet_delete',
    'ip_subnet_find_free': 'ip_find_free_subnet',
})

SERVICE_MAPPER.update({
    'ip_subnet6_create': 'ip6_subnet6_add',
    'ip_subnet6_update': 'ip6_subnet6_add',
    'ip_subnet6_count': 'ip6_block6_subnet6_count',
    'ip_subnet6_list': 'ip6_block6_subnet6_list',
    'ip_subnet6_info': 'ip6_block6_subnet6_info',
    'ip_subnet6_delete': 'ip6_subnet6_delete',
    'ip_subnet6_find_free': 'ip6_find_free_subnet6',
})

SERVICE_MAPPER.update({
    'ip_pool_create': 'ip_pool_add',
    'ip_pool_update': 'ip_pool_add',
    'ip_pool_count': 'ip_pool_count',
    'ip_pool_list': 'ip_pool_list',
    'ip_pool_info': 'ip_pool_info',
    'ip_pool_delete': 'ip_pool_delete',
})

SERVICE_MAPPER.update({
    'ip_pool6_create': 'ip6_pool6_add',
    'ip_pool6_update': 'ip6_pool6_add',
    'ip_pool6_count': 'ip6_pool6_count',
    'ip_pool6_list': 'ip6_pool6_list',
    'ip_pool6_info': 'ip6_pool6_info',
    'ip_pool6_delete': 'ip6_pool6_delete',
})

SERVICE_MAPPER.update({
    'ip_address_create': 'ip_add',
    'ip_address_update': 'ip_add',
    'ip_address_count': 'ip_address_count',
    'ip_address_list': 'ip_address_list',
    'ip_address_info': 'ip_address_info',
    'ip_address_delete': 'ip_delete',
    'ip_address_find_free': 'ip_find_free_address',
})

SERVICE_MAPPER.update({
    'ip_address6_create': 'ip6_address6_add',
    'ip_address6_update': 'ip6_address6_add',
    'ip_address6_count': 'ip6_address6_count',
    'ip_address6_list': 'ip6_address6_list',
    'ip_address6_info': 'ip6_address6_info',
    'ip_address6_delete': 'ip6_address6_delete',
    'ip_address6_find_free': 'ip6_find_free_address6',
})

SERVICE_MAPPER.update({
    'ip_alias_create': 'ip_alias_add',
    'ip_alias_update': 'ip_alias_add',
    'ip_alias_list': 'ip_alias_list',
    'ip_alias_delete': 'ip_alias_delete',
})

SERVICE_MAPPER.update({
    'ip_alias6_create': 'ip6_alias_add',
    'ip_alias6_update': 'ip6_alias_add',
    'ip_alias6_list': 'ip6_alias_list',
    'ip_alias6_delete': 'ip6_alias_delete',
})

SERVICE_MAPPER.update({
    # DNS
    'dns_server_create': 'dns_add',
    'dns_server_update': 'dns_add',
    'dns_server_delete': 'dns_delete',
    'dns_server_list': 'dns_server_list',
    'dns_server_info': 'dns_server_info',
    'dns_server_count': 'dns_server_count',

    'dns_view_create': 'dns_view_add',
    'dns_view_update': 'dns_view_add',
    'dns_view_list': 'dns_view_list',
    'dns_view_info': 'dns_view_info',
    'dns_view_count': 'dns_view_count',
    'dns_view_delete': 'dns_view_count',

    'dns_view_param_create': 'dns_view_param_add',
    'dns_view_param_update': 'dns_view_param_add',
    'dns_view_param_list': 'dns_view_param_list',
    'dns_view_param_info': 'dns_view_param_info',
    'dns_view_param_count': 'dns_view_param_count',
    'dns_view_param_delete': 'dns_view_param_delete',

    'dns_zone_create': 'dns_zone_add',
    'dns_zone_update': 'dns_zone_add',
    'dns_zone_delete': 'dns_zone_delete',
    'dns_zone_list': 'dns_zone_list',
    'dns_zone_info': 'dns_zone_info',
    'dns_zone_count': 'dns_zone_count',

    'dns_zone_param_create': 'dns_zone_param_add',
    'dns_zone_param_update': 'dns_zone_param_add',
    'dns_zone_param_delete': 'dns_zone_param_delete',
    'dns_zone_param_list': 'dns_zone_param_list',
    'dns_zone_param_info': 'dns_zone_param_info',
    'dns_zone_param_count': 'dns_zone_param_count',

    'dns_rr_list': 'dns_rr_list',
    'dns_rr_create': 'dns_rr_add',
    'dns_rr_info': 'dns_rr_info',
    'dns_rr_count': 'dns_rr_count',
    'dns_rr_update': 'dns_rr_add',
    'dns_rr_delete': 'dns_rr_delete',

    'dns_acl_create': 'dns_acl_add',
    'dns_acl_update': 'dns_acl_add',
    'dns_acl_delete': 'dns_acl_delete',
    'dns_acl_list': 'dns_acl_list',
    'dns_acl_info': 'dns_acl_info',
    'dns_acl_count': 'dns_acl_count',

    'dns_key_create': 'dns_key_add',
    'dns_key_update': 'dns_key_add',
    'dns_key_delete': 'dns_key_delete',
    'dns_key_list': 'dns_key_list',
    'dns_key_info': 'dns_key_info',
    'dns_key_count': 'dns_key_count',
})

SERVICE_MAPPER.update({
    # Application manager
    'app_application_list': 'app_application_list',
    'app_application_create': 'app_application_add',
    'app_application_update': 'app_application_add',
    'app_application_delete': 'app_application_delete',
    'app_application_count': 'app_application_count',
    'app_application_info': 'app_application_info',

    'app_pool_create': 'app_pool_add',
    'app_pool_update': 'app_pool_add',
    'app_pool_list': 'app_pool_list',
    'app_pool_count': 'app_pool_count',
    'app_pool_info': 'app_pool_info',
    'app_pool_delete': 'app_pool_delete',

    'app_node_create': 'app_node_add',
    'app_node_update': 'app_node_add',
    'app_node_info': 'app_node_info',
    'app_node_count': 'app_node_count',
    'app_node_list': 'app_node_list',
    'app_node_delete': 'app_node_delete',

    'app_healthcheck_count': 'app_healthcheck_count',
    'app_healthcheck_info': 'app_healthcheck_info',
    'app_healthcheck_list': 'app_healthcheck_list',
})

SERVICE_MAPPER.update({
    'member_list': 'member_list',
})

SERVICE_MAPPER.update({
    # dhcp
    'dhcp_server_create': 'dhcp_add',
    'dhcp_server_update': 'dhcp_add',
    'dhcp_server_delete': 'dhcp_delete',
    'dhcp_server_info': 'dhcp_server_info',
    'dhcp_server_count': 'dhcp_server_count',
    'dhcp_server_list': 'dhcp_server_list',

    'dhcp_scope_create': 'dhcp_scope_add',
    'dhcp_scope_update': 'dhcp_scope_add',
    'dhcp_scope_count': 'dhcp_scope_count',
    'dhcp_scope_list': 'dhcp_scope_list',
    'dhcp_scope_info': 'dhcp_scope_info',
    'dhcp_scope_delete': 'dhcp_scope_delete',

    'dhcp_shared_network_create': 'dhcp_sn_add',
    'dhcp_shared_network_update': 'dhcp_sn_add',
    'dhcp_shared_network_count': 'dhcp_shared_network_count',
    'dhcp_shared_network_list': 'dhcp_shared_network_list',
    'dhcp_shared_network_info': 'dhcp_shared_network_info',

    'dhcp_range_create': 'dhcp_range_add',
    'dhcp_range_update': 'dhcp_range_add',
    'dhcp_range_list': 'dhcp_range_list',
    'dhcp_range_info': 'dhcp_range_info',
    'dhcp_range_count': 'dhcp_range_count',
    'dhcp_range_delete': 'dhcp_range_delete',

    'dhcp_static_create': 'dhcp_static_add',
    'dhcp_static_update': 'dhcp_static_add',
    'dhcp_static_list': 'dhcp_static_list',
    'dhcp_static_info': 'dhcp_static_info',
    'dhcp_static_count': 'dhcp_static_count',
    'dhcp_static_delete': 'dhcp_static_delete',

    'dhcp_group_create': 'dhcp_group_add',
    'dhcp_group_update': 'dhcp_group_add',
    'dhcp_group_list': 'dhcp_group_list',
    'dhcp_group_info': 'dhcp_group_info',
    'dhcp_group_count': 'dhcp_group_count',
    'dhcp_group_delete': 'dhcp_group_delete'
})

SERVICE_MAPPER.update({
    # dhcp6
    'dhcp_server6_create': 'dhcp6_add',
    'dhcp_server6_update': 'dhcp6_add',
    'dhcp_server6_delete': 'dhcp6_delete',
    'dhcp_server6_info': 'dhcp6_server6_info',
    'dhcp_server6_count': 'dhcp6_server6_count',
    'dhcp_server6_list': 'dhcp6_server6_list',

    'dhcp_scope6_create': 'dhcp6_scope6_add',
    'dhcp_scope6_update': 'dhcp6_scope6_add',
    'dhcp_scope6_count': 'dhcp6_scope6_count',
    'dhcp_scope6_list': 'dhcp6_scope6_list',
    'dhcp_scope6_info': 'dhcp6_scope6_info',
    'dhcp_scope6_delete': 'dhcp6_scope6_delete',

    'dhcp_range6_create': 'dhcp6_range6_add',
    'dhcp_range6_update': 'dhcp6_range6_add',
    'dhcp_range6_list': 'dhcp6_range6_list',
    'dhcp_range6_info': 'dhcp6_range6_info',
    'dhcp_range6_count': 'dhcp6_range6_count',
    'dhcp_range6_delete': 'dhcp6_range6_delete',

    'dhcp_static6_create': 'dhcp6_static6_add',
    'dhcp_static6_update': 'dhcp6_static6_add',
    'dhcp_static6_list': 'dhcp6_static6_list',
    'dhcp_static6_info': 'dhcp6_static6_info',
    'dhcp_static6_count': 'dhcp6_static6_count',
    'dhcp_static6_delete': 'dhcp6_static6_delete',

})

SERVICE_MAPPER.update({
    # device manager
    'host_device_create': 'hostdev_add',
    'host_device_update': 'hostdev_add',
    'host_device_delete': 'hostdev_delete',
    'host_device_list':   'hostdev_list',
    'host_device_count':  'hostdev_count',
    'host_device_info':   'hostdev_info',

    'host_iface_create': 'hostiface_add',
    'host_iface_update': 'hostiface_add',
    'host_iface_delete': 'hostiface_delete',
    'host_iface_list':   'hostiface_list',
    'host_iface_count':  'hostiface_count',
    'host_iface_info':   'hostiface_info',

    'host_link_create':  'link_hostiface_add',
    'host_link_update':  'link_hostiface_add',
    'host_link_delete':  'link_hostiface_delete',
    'host_link_count':   'link_hostiface_count',
    'host_link_list':    'link_hostiface_list',
    'host_link_info':    'link_hostiface_list',
})

SERVICE_MAPPER.update({
    # Vlan Manager
    'vlm_domain_create':  'vlm_domain_add',
    'vlm_domain_update':  'vlm_domain_add',
    'vlm_domain_delete':  'vlm_domain_delete',
    'vlm_domain_count':   'vlmdomain_count',
    'vlm_domain_list':    'vlmdomain_list',
    'vlm_domain_info':    'vlmdomain_info',

    'vlm_range_create':  'vlm_range_add',
    'vlm_range_update':  'vlm_range_add',
    'vlm_range_delete':  'vlm_range_delete',
    'vlm_range_count':   'vlmrange_count',
    'vlm_range_list':    'vlmrange_list',
    'vlm_range_info':    'vlmrange_info',

    'vlm_vlan_create':  'vlm_vlan_add',
    'vlm_vlan_update':  'vlm_vlan_add',
    'vlm_vlan_delete':  'vlm_vlan_delete',
    'vlm_vlan_count':   'vlmvlan_count',
    'vlm_vlan_list':    'vlmvlan_list',
    'vlm_vlan_info':    'vlmvlan_info',
})

SERVICE_MAPPER.update({
    # Guardian params
    'guardian_param_create':  'guardian_param_add',
    'guardian_param_update':  'guardian_param_add',
    'guardian_param_count':   'guardian_param_count',
    'guardian_param_list':    'guardian_param_list',
    'guardian_param_info':    'guardian_param_info',
})

METHOD_MAPPER = {
    'add': 'POST',
    'update': 'PUT',
    'count': 'GET',
    'list': 'GET',
    'info': 'GET',
    'find_free': "OPTIONS",
    'create': 'POST',
    'delete': "DELETE",
}



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2020-07-26 16:20:12 alex>
#

"""
SOLIDserver device manager

"""

# import ipaddress
# import logging
# import pprint

from SOLIDserverRest.Exception import SDSError, SDSEmptyError
from SOLIDserverRest.Exception import SDSDeviceError, SDSDeviceNotFoundError

from .class_params import ClassParams

__all__ = ["Device"]


class Device(ClassParams):
    """ class to manipulate the SOLIDserver device """

    # -------------------------------------
    def __init__(self,
                 sds=None,
                 name=None,
                 class_params=None):
        """init a device object:
        - sds: object SOLIDserver, could be set afterwards
        - name: name of the device
        """

        if name is None:
            raise SDSDeviceError("no name provided at device init")

        super(Device, self).__init__(sds, name)

        # params mapping the object in SDS
        self.clean_params()

        self.set_sds(sds)
        self.set_name(name)
        self.aifs_simple = []
        self.ainterfaces = []

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    def clean_params(self):
        """ clean the object params """
        super(Device, self).clean_params()

        self.params = {
            'hostdev_id': None,
            'hostdev_name': None,
            'hostdev_class_name': None,
            'iplnetdev_id': None,
            'row_enabled': None,
            'iplnetdev_name': None,
            'hostdev_ip_addr': None,
            'hostdev_ip_formated': None,
            'hostdev_site_id': None,
            'hostdev_site_name': '',
            'port_total': None,
            'port_used': None,
            'port_used_percent': None,
            'port_free': None,
            'iface_total': None,
            'iface_used': None,
            'iface_used_percent': None,
            'iface_free': None
        }

        self.aifs_simple = []
        self.ainterfaces = []

    # -------------------------------------
    def set_param(self, param=None, value=None, exclude=None, name=None):
        super(Device, self).set_param(param,
                                      value,
                                      exclude=['hostdev_id'],
                                      name='hostdev_name')

    # -------------------------------------
    def create(self):
        """ create the device in SDS """

        if self.sds is None:
            raise SDSDeviceError(message="not connected")

        params = {
            'hostdev_name': self.name,
            **self.additional_params
        }

        if 'hostdev_class_name' in self.params:
            self.add_class_params({
                'hostdev_class_name': self.params['hostdev_class_name']
            })
            params['hostdev_class_name'] = self.params['hostdev_class_name']

        self.prepare_class_params('hostdev', params)

        rjson = self.sds.query("host_device_create",
                               params=params)

        if 'errmsg' in rjson:
            raise SDSDeviceError(message="dev creation error, "
                                 + rjson['errmsg'])

        self.params['hostdev_id'] = int(rjson[0]['ret_oid'])
        self.myid = int(self.params['hostdev_id'])

        self.refresh()

    # -------------------------------------
    def update(self):
        """ update the device in SDS """

        if self.sds is None:
            raise SDSDeviceError(message="not connected")

        params = {
            'hostdev_id': self._get_id(query="host_device_list",
                                       key="hostdev"),
            'hostdev_name': self.name,
            **self.additional_params
        }

        if 'hostdev_class_name' in self.params:
            self.add_class_params({
                'hostdev_class_name': self.params['hostdev_class_name']
            })
            params['hostdev_class_name'] = self.params['hostdev_class_name']

        self.prepare_class_params('hostdev', params)

        rjson = self.sds.query("host_device_update",
                               params=params)

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSDeviceError(message="device update error, "
                                 + rjson['errmsg'])

        self.refresh()

    # -------------------------------------
    def delete(self):
        """deletes the device in the SDS"""
        if self.sds is None:
            raise SDSDeviceError(message="not connected")

        if self.params['hostdev_id'] is None:
            raise SDSDeviceNotFoundError("on delete")

        params = {
            'hostdev_id': self.params['hostdev_id'],
            **self.additional_params
        }

        try:
            rjson = self.sds.query("host_device_delete",
                                   params=params)
        except SDSEmptyError:
            raise SDSDeviceNotFoundError(message="not found")

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSDeviceNotFoundError("on delete "+rjson['errmsg'])

        self.clean_params()

    # -------------------------------------
    def refresh(self):
        """refresh content of the device from the SDS"""

        if self.sds is None:
            raise SDSDeviceError(message="not connected")

        device_id = self._get_id(query="host_device_list",
                                 key="hostdev")

        params = {
            "hostdev_id": device_id,
            **self.additional_params
        }

        try:
            rjson = self.sds.query("host_device_info",
                                   params=params)
        except SDSError as err_descr:
            msg = "cannot get device info on id={}".format(device_id)
            msg += " / "+str(err_descr)
            raise SDSDeviceError(msg)

        rjson = rjson[0]

        for label in ['hostdev_id',
                      'hostdev_name',
                      'hostdev_class_name',
                      'iplnetdev_id',
                      'row_enabled',
                      'iplnetdev_name',
                      'hostdev_ip_addr',
                      'hostdev_ip_formated',
                      'hostdev_site_id',
                      'hostdev_site_name',
                      'port_total',
                      'port_used',
                      'port_used_percent',
                      'port_free',
                      'iface_total',
                      'iface_used',
                      'iface_used_percent',
                      'iface_free']:
            if label not in rjson:  # pragma: no cover
                msg = "parameter {} not found in device".format(label)
                raise SDSDeviceError(msg)
            self.params[label] = rjson[label]

        if 'hostdev_class_parameters' in rjson:
            self.update_class_params(rjson['hostdev_class_parameters'])

    # -------------------------------------
    def fetch_interfaces(self):
        """ fetch interfaces if any"""

        if self.sds is None:  # pragma: no cover
            raise SDSDeviceError(message="not connected")

        if self.myid == -1:  # pragma: no cover
            raise SDSDeviceError(message="device refresh needed")

        params = {
            **self.additional_params,
        }

        params['WHERE'] = "hostdev_id='{}'".format(self.myid)

        rjson = self.sds.query("host_iface_list",
                               params=params)

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSDeviceError(message="interface list error, "
                                 + rjson['errmsg'])

        self.aifs_simple = []
        for interface in rjson:
            self.aifs_simple.append({
                'if': interface['hostiface_name'],
                'id': int(interface['hostiface_id'])
            })

    # -------------------------------------
    def link_interface(self, interface):
        """link back an interface object to the device"""
        if interface not in self.ainterfaces:
            self.ainterfaces.append(interface)

    # -------------------------------------
    def __str__(self):  # pragma: no cover
        """return the string notation of the device object"""

        return_val = "*device* name={}".format(self.name)

        if ('hostdev_site_name' in self.params
                and self.params['hostdev_site_name'] != ''):
            return_val += " space={}".format(self.params['hostdev_site_name'])
            return_val += " [{}]".format(self.params['hostdev_site_id'])

        return_val += self.str_params(exclude=['hostdev_id',
                                               'hostdev_name'])

        if self.aifs_simple:
            return_val += " interfaces=["
            sep = ""
            for intf in self.aifs_simple:
                return_val += sep + intf['if']
            return_val += "]"

        return_val += str(super(Device, self).__str__())

        return return_val



#!/usr/bin/python
# -*-coding:Utf-8 -*
#
# disable naming convention issue
# pylint: disable=C0103
##########################################################
"""
Exceptions for the SOLIDServer modules
"""

# import logging

__all__ = [
    "SDSError",

    "SDSAuthError",
    "SDSEmptyError",
    "SDSInitError",
    "SDSRequestError",
    "SDSServiceError",

    "SDSSpaceError",

    "SDSDeviceError", "SDSDeviceNotFoundError",
    "SDSDeviceIfError", "SDSDeviceIfNotFoundError",
    "SDSNetworkError", "SDSNetworkNotFoundError",
    "SDSIpAddressError", "SDSIpAddressNotFoundError",

    "SDSDNSError", "SDSDNSAlreadyExistingError",
    "SDSDNSCredentialsError",
]


# --------------------------------------------------------------------------
class SDSError(Exception):
    """ generic class for any exception in SOLIDServer communication """

    def __init__(self, message=""):
        super(SDSError, self).__init__()
        self.message = message

    def __str__(self):
        return "{}".format(self.message)


class SDSInitError(SDSError):
    """ raised when action on non initialized SSD connection """

    def __init__(self, message=""):
        super(SDSInitError, self).__init__("[init] {}".format(message))


class SDSServiceError(SDSError):
    """ raised on unknown service """

    def __init__(self, service_name, message=""):
        super(SDSServiceError, self).__init__(message)
        self.service = service_name

    def __str__(self):
        return "{} on service {}".format(self.message, self.service)


class SDSRequestError(SDSError):
    """ raised when urllib request is failing """

    def __init__(self, method, url, headers, message=""):
        super(SDSRequestError, self).__init__(message)
        self.method = method
        self.url = url
        self.headers = headers

    def __str__(self):
        return "{} with {} {}".format(self.message, self.method, self.url)


# this class cannot be tests with non connected coverage
class SDSAuthError(SDSError):   # pragma: no cover
    """ raised when auth on request is wrong """

    def __init__(self, message=""):
        super(SDSAuthError, self).__init__("authent: {}".format(message))


# this class cannot be tests with non connected coverage
class SDSEmptyError(SDSError):   # pragma: no cover
    """ raised when empty answer """

    def __init__(self, message=""):
        super(SDSEmptyError, self).__init__("empty answer: {}".format(message))


class SDSSpaceError(SDSError):   # pragma: no cover
    """ raised when error on space """

    def __init__(self, message=""):
        super(SDSSpaceError, self).__init__("space error: {}".format(message))


class SDSDeviceError(SDSError):   # pragma: no cover
    """ raised when error on device """

    def __init__(self, message=""):
        message = "device error: {}".format(message)
        super(SDSDeviceError, self).__init__(message)


class SDSDeviceNotFoundError(SDSError):   # pragma: no cover
    """ raised when device not found """

    def __init__(self, message=""):
        message = "device not found: {}".format(message)
        super(SDSDeviceNotFoundError, self).__init__(message)


class SDSDeviceIfError(SDSError):   # pragma: no cover
    """ raised when error on device interface """

    def __init__(self, message=""):
        message = "device interface error: {}".format(message)
        super(SDSDeviceIfError, self).__init__(message)


class SDSDeviceIfNotFoundError(SDSError):   # pragma: no cover
    """ raised when device interface not found """

    def __init__(self, message=""):
        message = "device interface not found: {}".format(message)
        super(SDSDeviceIfNotFoundError, self).__init__(message)


class SDSNetworkError(SDSError):   # pragma: no cover
    """ raised when error on network """

    def __init__(self, message=""):
        message = "network error: {}".format(message)
        super(SDSNetworkError, self).__init__(message)


class SDSNetworkNotFoundError(SDSError):   # pragma: no cover
    """ raised when network not found """

    def __init__(self, message=""):
        message = "network not found: {}".format(message)
        super(SDSNetworkNotFoundError, self).__init__(message)


class SDSIpAddressError(SDSError):   # pragma: no cover
    """ raised when error on ip address """

    def __init__(self, message=""):
        message = "ip error: {}".format(message)
        super(SDSIpAddressError, self).__init__(message)


class SDSIpAddressNotFoundError(SDSError):   # pragma: no cover
    """ raised when ip address not found """

    def __init__(self, message=""):
        message = "ip not found: {}".format(message)
        super(SDSIpAddressNotFoundError, self).__init__(message)


class SDSDNSError(SDSError):   # pragma: no cover
    """ raised when dns server error """

    def __init__(self, message=""):
        message = "DNS server error: {}".format(message)
        super(SDSDNSError, self).__init__(message)


class SDSDNSAlreadyExistingError(SDSDNSError):   # pragma: no cover
    """ raised when dns server is already existing """

    def __init__(self, message=""):
        message = "DNS server already existing: {}".format(message)
        super(SDSDNSAlreadyExistingError, self).__init__(message)


class SDSDNSCredentialsError(SDSDNSError):   # pragma: no cover
    """ raised on dns server connexion error  """

    def __init__(self, message=""):
        message = "DNS server credentials: {}".format(message)
        super(SDSDNSCredentialsError, self).__init__(message)



#
# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-08-25 15:36:19 alex>
#

"""
SOLIDserver DNS zone management

"""

# import ipaddress
import logging
import time

from SOLIDserverRest.Exception import (SDSInitError,
                                       SDSError,
                                       SDSEmptyError,
                                       SDSDNSError)

from .class_params import ClassParams
from .dns import DNS
from .sds import SDS


class DNS_zone(ClassParams):  # pylint: disable=C0103
    """ class to manipulate a DNS zone object """
    TYPE_MASTER = "master"
    TYPE_SLAVE = "slave"
    TYPE_HINT = "hint"
    TYPE_STUB = "stub"
    TYPE_FWD = "forward"
    TYPE_DELONLY = "delegation-only"
    TYPE_CLOUD = "cloud"

    TYPES_LIST = [TYPE_SLAVE,
                  TYPE_MASTER,
                  TYPE_HINT,
                  TYPE_STUB,
                  TYPE_FWD,
                  TYPE_DELONLY,
                  TYPE_CLOUD]

    # -------------------------------------

    def __init__(self, sds=None,
                 name=None,
                 zone_type=None,
                 class_params=None):
        """init the DNS zone object:
        - sds: object SOLIDserver, could be set afterwards
        - name: dns zone name
        - zone_type: dns zone type (master, slave, hint,
                                    stub, forward, delegation-only,
                                    cloud)
        - zone_resolution: ?
        """

        if sds and not isinstance(sds, SDS):
            raise SDSInitError(message="sds param is not of type SDS")

        super(DNS_zone, self).__init__(sds, name)

        self.zone_type = None
        self.dns_server = None

        if zone_type:
            # TODO add checks  pylint: disable=W0511
            if zone_type not in self.TYPES_LIST:
                raise SDSDNSError(message="zone type unknown")
            self.zone_type = zone_type

        self.params = {
            'dnszone_is_reverse': 0
        }

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    def set_dns(self, dns):
        """ link the zone to a dns server object """
        if not isinstance(dns, DNS):
            raise SDSDNSError(message="zone for linking is not a zone object")

        self.dns_server = dns

    # -------------------------------------
    def set_type(self, zone_type):
        """ set the type for the zone """

        if str(zone_type) not in self.TYPES_LIST:
            raise SDSDNSError(message="zone type not supported")

        self.zone_type = str(zone_type)

    # -------------------------------------
    def set_is_reverse(self, reverse_flag):
        """ set if the zone is reverse """

        if reverse_flag:
            self.params['dnszone_is_reverse'] = 1

    # -------------------------------------
    def create(self, sync=True):
        """creates the DNS zone"""
        if self.sds is None:
            raise SDSInitError(message="not connected")

        if self.dns_server is None:
            raise SDSDNSError(message="zone not linked to a DNS server")

        if self.zone_type not in self.TYPES_LIST:
            raise SDSDNSError(message="zone type not set")

        params = {
            'dnszone_name': self.name,
            'dnszone_type': self.zone_type,
            'dns_id': self.dns_server.myid,
            **self.additional_params
        }

        self.prepare_class_params('dnszone', params)

        try:
            rjson = self.sds.query("dns_zone_create",
                                   params=params)
        except SDSError:   # pragma: no cover
            raise SDSDNSError(message="create DNS zone")

        if 'errno' in rjson:
            raise SDSDNSError(message="create DNS zone"
                              " {}".format(rjson['errmsg']))

        rjson = rjson[0]
        if 'ret_oid' in rjson:
            self.myid = int(rjson['ret_oid'])

        if sync:
            time.sleep(.5)
            self.refresh()

    # -------------------------------------
    def _get_zoneid_by_name(self, name):
        """get the DNS zone ID from its name, return None if non existant"""

        try:
            rjson = self.sds.query("dns_zone_list",
                                   params={
                                       "WHERE": "dnszone_name='{}'".
                                                format(name),
                                       **self.additional_params
                                   })
        except SDSEmptyError:
            return None

        if rjson[0]['errno'] != '0':
            raise SDSDNSError("dnszone_id errno raised")

        return rjson[0]['dnszone_id']

    # -------------------------------------
    def _wait_for_synch(self, delete=False):
        """waith for the DNS zone to be in sync"""
        if self.myid is None or self.myid == -1:
            raise SDSDNSError(message="missing DNS id")

        _wait_delay = 2.0

        for _ in range(10):
            try:
                rjson = self.sds.query("dns_zone_info",
                                       params={
                                           "dnszone_id": self.myid,
                                       })
            except SDSError:
                return None

            if not rjson:   # pragma: no cover
                raise SDSDNSError(message="DNS zone sync error")

            if not delete:
                # we wait for the zone to be pushed to the server
                if rjson[0]['delayed_create_time'] == '0':
                    return rjson
            else:
                # we wait for the zone to be deleted from the server
                if rjson[0]['delayed_delete_time'] == '0':
                    return None

            logging.debug('not yet in synch %d', _wait_delay)
            time.sleep(_wait_delay)
            _wait_delay *= 1.2

        raise SDSDNSError(message="DNS zone sync takes too long")

    # -------------------------------------
    def delete(self, sync=True):
        """deletes the DNS zone from the server"""
        if self.sds is None:
            raise SDSDNSError(message="not connected")

        if self.myid is None or self.myid == -1:
            raise SDSDNSError(message="missing DNS id")

        try:
            logging.debug("delete")
            rjson = self.sds.query("dns_zone_delete",
                                   params={
                                       'dnszone_id': self.myid,
                                       **self.additional_params
                                   })
            if 'errmsg' in rjson:  # pragma: no cover
                raise SDSDNSError(message="DNS zone delete error, "
                                  + rjson['errmsg'])
        except SDSError:
            raise SDSDNSError(message="DNS zone delete error")

        if sync:
            time.sleep(2)
            self._wait_for_synch(delete=True)

        self.myid = -1

    # -------------------------------------
    def refresh(self):
        """refresh content of the DNS zone from the SDS"""
        if self.sds is None:
            raise SDSDNSError(message="not connected")

        if self.dns_server is None:
            raise SDSDNSError(message="zone not linked to a server")

        if self.myid is None or self.myid == -1:
            zone_id = self._get_zoneid_by_name(self.name)
        else:
            zone_id = self.myid

        if zone_id is None:
            raise SDSDNSError(message="non existant DNS zone to refresh")

        self.myid = zone_id
        rjson = self._wait_for_synch()

        if not rjson:   # pragma: no cover
            raise SDSDNSError(message="DNS server refresh error, len of array")

        rjson = rjson[0]

        for label in [
                'dns_state',
                'dnszone_ad_integrated',
                'dnszone_allow_query',
                'dnszone_allow_transfer',
                'dnszone_allow_update',
                'dnszone_also_notify',
                'dnszone_class_name',
                'dnszone_forward',
                'dnszone_forwarders',
                'dnszone_is_reverse',
                'dnszone_is_rpz',
                'dnszone_masters',
                'dnszone_name',
                'dnszone_name_utf',
                'dnszone_notify',
                'dnszone_order',
                'dnszone_rev_sort_zone',
                'dnszone_type',
                'ds',
                'num_keys',
        ]:
            if label not in rjson:   # pragma: no cover
                raise SDSDNSError("parameter"
                                  + " {}".format(label)
                                  + " not found in DNS zone")
            self.params[label] = rjson[label]

        if 'dnszone_class_parameters' in rjson:
            self.update_class_params(rjson['dnszone_class_parameters'])

        self.zone_type = rjson['dnszone_type']

    # -------------------------------------
    def __str__(self):
        """return the string notation of the DNS zone object"""
        return_val = "*ZONE* name={}".format(self.name)

        if self.myid and self.myid != -1:
            return_val += " [#{}]".format(self.myid)

        if self.dns_server:
            return_val += " server={}".format(self.dns_server.name)

        if 'dnszone_is_reverse' in self.params:
            if self.params['dnszone_is_reverse'] == '1':
                return_val += " resolve=reverse"
            else:
                return_val += " resolve=name"
        else:
            return_val += " not refreshed"

        return_val += str(super(DNS_zone, self).__str__())

        return return_val



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-05-12 09:43:16 alex>
#
# pylint: disable=R0801


"""
SOLIDserver network manager

"""

# pylint: disable=too-many-branches

import math

from packaging.version import Version, parse
from SOLIDserverRest.Exception import (SDSEmptyError, SDSError,
                                       SDSNetworkError,
                                       SDSNetworkNotFoundError)

from .class_params import ClassParams
from .space import Space

# import logging
# import pprint


# pylint: disable=R0902
class Network(ClassParams):
    """ class to manipulate the SOLIDserver network """

    # -------------------------------------
    def __init__(self, sds=None,  # pylint: disable=too-many-arguments
                 space=None,
                 name=None,
                 class_params=None):
        """init a network object:
        - sds: object SOLIDserver, could be set afterwards
        - space: space object for this network
        - name: name of the subnet
        """

        super().__init__(sds, name)

        # params mapping the object in SDS
        self.clean_params()

        # reset the name since suppressed by the clean proc
        self.name = name

        self.set_sds(sds)
        # self.set_name(name)

        self.description = None

        if space and not isinstance(space, Space):
            raise SDSNetworkError("no valid space provided")
        self.space = space

        self.subnet_addr = None
        self.subnet_prefix = None
        self.is_block = False
        self.is_terminal = False
        self.parent_network = None

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    def clean_params(self):
        """ clean the object params """

        super().clean_params()

        self.subnet_addr = None
        self.subnet_prefix = None

        self.is_block = False
        self.is_terminal = False

        self.parent_network = None

        self.params = {
            'subnet_id': None,
            'parent_subnet_id': None
        }

    # -------------------------------------
    def set_address_prefix(self, ipaddress, prefix):
        """set the address and prefix of this network"""
        # need to normalize and check the ip address
        self.subnet_addr = ipaddress
        self.subnet_prefix = prefix

        if self.in_sync:  # pragma: no cover
            self.update()

    # -------------------------------------
    def set_is_block(self, block=False):
        """is this network a block"""
        self.is_block = block
        if block:
            self.set_is_terminal(False)

        if self.in_sync:  # pragma: no cover
            self.update()

    # -------------------------------------
    def set_is_terminal(self, terminal=False):
        """is this network a terminal"""
        self.is_terminal = terminal
        if terminal:
            self.set_is_block(False)

        if self.in_sync:  # pragma: no cover
            self.update()

    # -------------------------------------
    def set_parent(self, network):
        """set the parent network => not a block then"""
        if network.myid == -1:
            raise SDSNetworkError("no valid parent network found")

        self.parent_network = network
        self.set_is_block(False)

        if self.in_sync:  # pragma: no cover
            self.update()

    # -------------------------------------
    def set_param(self, param=None, value=None, exclude=None, name=None):
        """ set a specific param on the network object """
        if param == 'description':
            self.description = str(value)
            self.set_class_params({'__eip_description': self.description})
            return

        super().set_param(param,
                          value,
                          exclude=['subnet_id'],
                          name='subnet_name')

    # -------------------------------------
    def find_free(self, prefix, max_find=4):
        """ find the next free subnets in the space within this network
            by order of priority to avoid fragmentation
        """
        params = {
            'site_id': self.space.params['site_id'],
            'prefix': prefix,
            'max_find': max_find,
            'begin_addr': self.subnet_addr,
            **self.additional_params
        }

        params['WHERE'] = 'block_id={}'.format(self.myid)

        try:
            rjson = self.sds.query("ip_subnet_find_free",
                                   params=params)
        except SDSEmptyError:
            return None

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSNetworkError(message="find free net, " +
                                  rjson['errmsg'])

        aip = []
        for net in rjson:
            iphex = net['start_ip_addr']
            ipv4_addr = "{}.{}.{}.{}".format(int(iphex[0:2], 16),
                                             int(iphex[2:4], 16),
                                             int(iphex[4:6], 16),
                                             int(iphex[6:8], 16))
            aip.append(ipv4_addr)

        return aip

    # -------------------------------------
    def find_free_ip(self, max_find=4):
        """ find the next free ip in the current subnet
        """
        params = {
            'max_find': max_find,
            'subnet_id': self.myid,
            **self.additional_params
        }

        try:
            rjson = self.sds.query("ip_address_find_free",
                                   params=params)
        except SDSEmptyError:
            return None

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSNetworkError(message="find free ip, " +
                                  rjson['errmsg'])

        aip = []
        for net in rjson:
            aip.append(net['hostaddr'])

        return aip

    # -------------------------------------
    def get_subnet_list(self,    # pylint: disable=too-many-arguments
                        depth=1,  # only one level
                        terminal=None,
                        offset=0,
                        page=25,
                        limit=50,
                        collected=0):
        """return the list of subnet in the parent subnet"""
        params = {
            'limit': page,
            'offset': offset,
            'ORDERBY': 'start_ip_addr',
            ** self.additional_params,
        }

        if limit > 0:
            if page > limit:
                params['limit'] = limit

        if 'WHERE' not in params:
            params['WHERE'] = ""
        else:
            params['WHERE'] += " and "

        params['WHERE'] += "site_id='{}'".format(self.space.params['site_id'])

        # look only under the block
        params['WHERE'] += " and subnet_path like '%#{}#%'".format(
            str(self.myid))

        if depth == 1:
            params['WHERE'] += "and parent_subnet_id='{}'".format(self.myid)

        if terminal is not None:
            if terminal in [1, 0]:
                params['WHERE'] += " and is_terminal='{}'".format(terminal)

        try:
            rjson = self.sds.query("ip_subnet_list",
                                   params=params)
        except SDSEmptyError:
            return None

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSNetworkError(message="net list, " +
                                  rjson['errmsg'])

        anets = []
        for net in rjson:
            anets.append({
                'start_hostaddr': net['start_hostaddr'],
                'subnet_size': 32 - int(math.log(int(net['subnet_size']), 2)),
                'subnet_name': net['subnet_name']
            })

        # no limit, we should get all the records
        if len(rjson) == page:
            if limit == 0 or collected < limit:
                newnets = self.get_subnet_list(depth, terminal,
                                               offset + page,
                                               page=page,
                                               limit=limit,
                                               collected=(len(anets) +
                                                          collected))
                if newnets is not None:
                    anets += newnets

        if limit and len(anets) > limit:
            anets = anets[:limit]

        return anets

    # -------------------------------------
    def create(self):
        """ create the subnet in SDS """

        if self.sds is None:
            raise SDSNetworkError(message="not connected")

        if self.space is None:
            raise SDSNetworkError("no space attached to network for create")

        if self.subnet_addr is None:
            raise SDSNetworkError("no address on network for create")

        if self.subnet_prefix is None:
            raise SDSNetworkError("no address size on network for create")

        # if object already created
        if self.myid > 0:
            return

        params = {
            'subnet_addr': self.subnet_addr,
            'subnet_prefix': self.subnet_prefix,
            'subnet_name': self.name,
            'site_id': self.space.params['site_id'],
            **self.additional_params
        }

        if self.is_block:
            params['is_terminal'] = '0'
            params['subnet_level'] = '0'
        else:
            if self.parent_network is not None:
                params['parent_subnet_id'] = self.parent_network.myid
            else:  # pragma: no cover
                assert None, "TODO - not a block and no parent set, abort"

            if self.is_terminal:
                params['is_terminal'] = '1'
            else:
                params['is_terminal'] = '0'

        self.prepare_class_params('network', params)

        rjson = self.sds.query("ip_subnet_create",
                               params=params)

        if 'errmsg' in rjson:
            raise SDSNetworkError(message="creation, " +
                                  rjson['errmsg'])

        self.params['subnet_id'] = int(rjson[0]['ret_oid'])
        self.myid = int(self.params['subnet_id'])

        self.refresh()

    # -------------------------------------
    def update(self):
        """ update the network in SDS """

        if self.sds is None:
            raise SDSNetworkError(message="not connected")

        params = {
            'subnet_id': self._get_id(query="ip_subnet_list",
                                      key="subnet"),
            'subnet_name': self.name,
            **self.additional_params
        }

        self.prepare_class_params('network', params)

        # logging.info(params)

        rjson = self.sds.query("ip_subnet_update",
                               params=params)

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSNetworkError(message="network update error, " +
                                  rjson['errmsg'])

        self.refresh()

    # -------------------------------------
    def delete(self):
        """deletes the network in the SDS"""
        if self.sds is None:
            raise SDSNetworkError(message="not connected")

        if self.params['subnet_id'] is None:
            raise SDSNetworkNotFoundError("on delete")

        params = {
            'subnet_id': self.params['subnet_id'],
            **self.additional_params
        }

        self.sds.query("ip_subnet_delete",
                       params=params)

        self.clean_params()

    # -------------------------------------
    def refresh(self):
        """refresh content of the network from the SDS"""

        if self.sds is None:
            raise SDSNetworkError(message="not connected")

        try:
            subnet_id = self._get_id(query="ip_subnet_list",
                                     key="subnet")
        except SDSError as err_descr:
            msg = "cannot get network id"
            msg += " / " + str(err_descr)
            raise SDSNetworkError(msg) from err_descr

        params = {
            "subnet_id": subnet_id,
            **self.additional_params
        }

        try:
            rjson = self.sds.query("ip_subnet_info",
                                   params=params)
        except SDSError as err_descr:
            msg = "cannot get network info on id={}".format(subnet_id)
            msg += " / " + str(err_descr)
            raise SDSNetworkError(msg) from err_descr

        rjson = rjson[0]
        # logging.info(rjson)

        labels = ['subnet_id',
                  'subnet_name',
                  'subnet_size',
                  'subnet_level',
                  'parent_subnet_id',
                  'is_terminal',
                  'subnet_allocated_size',
                  'subnet_allocated_percent',
                  'subnet_used_size',
                  'subnet_used_percent',
                  'subnet_ip_used_size',
                  'subnet_ip_used_percent',
                  'subnet_ip_free_size',
                  'is_in_orphan',
                  'lock_network_broadcast',
                  'tree_level']
        if parse(self.sds.get_version()) >= Version("7.0.0"):
            labels.extend(['start_hostaddr', 'end_hostaddr'])

        for label in labels:
            if label not in rjson:  # pragma: no cover
                msg = "parameter {} not found in network".format(label)
                raise SDSNetworkError(msg)
            self.params[label] = rjson[label]

        self.myid = int(rjson['subnet_id'])
        if rjson['is_terminal'] == '1':
            self.is_terminal = True
        else:
            self.is_terminal = False
            if rjson['subnet_level'] == '0':
                self.is_block = True
            else:
                self.is_block = False

        if 'start_hostaddr' in rjson:
            self.subnet_addr = rjson['start_hostaddr']

        # should be this variable (see API doc), but not working...
        if 'network_class_parameters' in rjson:   # pragma: no cover
            self.update_class_params(rjson['network_class_parameters'])

        if 'subnet_class_parameters' in rjson:
            self.update_class_params(rjson['subnet_class_parameters'])

        descr = self.get_class_params('__eip_description')
        if descr is not None:
            self.description = descr

    # -------------------------------------
    def __str__(self):  # pragma: no cover
        """return the string notation of the network object"""

        return_val = "*network* name={}".format(self.name)

        if self.description is not None:
            return_val += " \"{}\"".format(self.description)

        if self.is_block:
            return_val += " [block]"

        if self.is_terminal:
            return_val += " [terminal]"
        else:
            return_val += " [network]"

        return_val += self.str_params(exclude=['subnet_id',
                                               'subnet_name'])

        if self.parent_network:
            return_val += " parent={}".format(self.parent_network.myid)

        return_val += str(super().__str__())

        return return_val



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2020-04-13 16:10:32 alex>
#
# only for python v3

"""
SOLIDserver base object with class parameters
"""

import base64
import urllib
import logging

from .base import Base

__all__ = ["ClassParams"]


class ClassParams(Base):
    """ standard class for all objects in SDS with class parameters """
    # ---------------------------

    def __init__(self, sds=None, name=None):
        """init the object:
        """
        super().__init__(sds, name)

        self.fct_url_encode = urllib.parse.urlencode
        self.fct_b64_encode = base64.b64encode

        self.dclasses = {}
        self.__class_params = {}

        self.class_name = None

    # ---------------------------
    @classmethod
    def decode_class_params(cls, params, val):
        """push decoded parameters in the params structure"""
        if val == "":
            return None

        dir_val = urllib.parse.parse_qsl(val)

        params.update(dir_val)

        # specific
        if 'domain_list' in params:
            dlist = str.split(params['domain_list'], ';')
            params['domain_list'] = dlist

        return True

    # ---------------------------
    @classmethod
    def encode_class_params(cls, params):
        """get parameters from the structure and create string"""

        if not isinstance(params, dict):
            return None

        return urllib.parse.urlencode(params)

    # ---------------------------
    def get_class_params(self, key=None):
        """ get all/one class param """
        if key is None:
            return self.__class_params

        if not isinstance(key, str):
            logging.warning("get_class_params only accepting string as key")
            return None

        if key in self.__class_params:
            return self.__class_params[key]

        return None

    # ---------------------------
    def set_class_params(self, params=None):
        """ set the class param """
        if params is None:
            return None

        if not isinstance(params, dict):
            logging.warning("set class params only support dictionary")
            return None

        self.__class_params = params

        return True

    # ---------------------------
    def add_class_params(self, params=None):
        """ update the class param by adding this part """
        if params is None:
            return None

        if not isinstance(params, dict):
            logging.warning("update class params only support dictionary")
            return None

        self.__class_params.update(params)
        # logging.info(self.__class_params)

        return True

    # ---------------------------
    def prepare_class_params(self, keyprefix=None, params=None):
        """ encode the params into the string and update the dictionary """
        if keyprefix is None:
            return None

        if self.class_name:
            key = "{}_class_name".format(keyprefix)
            params[key] = self.class_name

        if params is None:
            return None

        if not isinstance(params, dict):
            return None

        if self.__class_params == {}:
            return None

        key = "{}_class_parameters".format(keyprefix)
        params[key] = self.encode_class_params(self.__class_params)

        # logging.info(params)

        return True

    # ---------------------------
    def update_class_params(self, params=None):
        """ update from a refresh """

        if params is None:
            return None

        if params == "":
            return None

        if isinstance(params, str):
            self.decode_class_params(self.__class_params,
                                     params)
            return True

        if isinstance(params, dict):
            self.__class_params.update(params)

        return True

    # -------------------------------------
    def set_class_name(self, name):
        """ set the class name for the object """
        self.class_name = name

    # -------------------------------------
    def __str__(self):  # pragma: no cover
        if self.__class_params == {}:
            return ""

        return_val = " cparams=["

        sep = ""
        for key, value in sorted(self.__class_params.items()):
            return_val += "{}{}={}".format(sep, key, value)
            sep = ", "

        return_val += "]"

        return_val += str(super().__str__())

        return return_val



#
# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-03-04 18:26:15 alex>
#

"""
SOLIDserver space management

get an existing space:
    space = sdsadv.Space(sds=sds, name="Local")
    space.refresh()

create a new space:
    space = sdsadv.Space(sds, name="test")
    space.create()

"""

import logging
# import math
# import pprint

from SOLIDserverRest.Exception import SDSInitError, SDSError
from SOLIDserverRest.Exception import SDSEmptyError, SDSSpaceError
# from SOLIDserverRest.Exception import SDSNetworkError

from .class_params import ClassParams


class Space(ClassParams):
    """ class to manipulate the SOLIDserver spaces """

    # -------------------------------------
    def __init__(self, sds=None,
                 name="Local",
                 class_params=None):
        """init the space object:
        - sds: object SOLIDserver, could be set afterwards
        - name: space name, default Local
        """

        super(Space, self).__init__(sds, name)

        # self.name = name
        # self.sds = sds

        self.params = {
            'site_is_template': None,
            'site_id': None,
            'tree_level': None,
            'site_name': None,
            'site_description': None,
            'parent_site_id': None,
            'parent_site_name': None,
            'site_class_name': None,
            'parent_site_class_name': None,
            'row_enabled': None,
            'multistatus': None,
        }

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    def create(self):
        """creates the space in the SDS"""
        if self.sds is None:
            raise SDSSpaceError(message="not connected")

        space_id = None
        try:
            space_id = self._get_id_by_name('ip_site_list', 'site', self.name)
        except SDSError:
            None  # pylint: disable=W0104

        # space_id = self._get_siteid_by_name(self.name)
        if space_id is not None:
            raise SDSSpaceError(message="already existant space")

        params = {
            'site_name': self.name,
            **self.additional_params
        }

        self.prepare_class_params('site', params)

        try:
            rjson = self.sds.query("ip_site_create",
                                   params=params)
        except SDSError:   # pragma: no cover
            logging.error("create space")

        if len(rjson) != 1:   # pragma: no cover
            raise SDSSpaceError(message="space creation error,"
                                + " array not recognized")
        if 'ret_oid' not in rjson[0]:   # pragma: no cover
            raise SDSSpaceError(message="space creation error, id not found")

        self.params['site_id'] = int(rjson[0]['ret_oid'])
        self.refresh()

    # -------------------------------------
    def delete(self):
        """deletes the space in the SDS"""
        if self.sds is None:
            raise SDSSpaceError(message="not connected")

        # space_id = self._get_siteid_by_name(self.name)
        space_id = self._get_id_by_name('ip_site_list', 'site', self.name)

        try:
            rjson = self.sds.query("ip_site_delete",
                                   params={
                                       'site_id': space_id,
                                       **self.additional_params
                                   })
            if 'errmsg' in rjson:  # pragma: no cover
                raise SDSSpaceError(message="space delete error, "
                                    + rjson['errmsg'])
        except SDSError:   # pragma: no cover
            raise SDSSpaceError(message="space delete error")

    # -------------------------------------
    def refresh(self):
        """refresh content of the object from the SDS"""
        if self.sds is None:
            raise SDSInitError(message="not connected")

        if self.params['site_id'] is None:
            space_id = self._get_id_by_name('ip_site_list', 'site', self.name)
        else:
            space_id = self.params['site_id']

        if space_id is None:
            raise SDSEmptyError(message="non existant space")

        rjson = self.sds.query("ip_site_info",
                               params={
                                   "site_id": space_id,
                                   **self.additional_params
                               })

        if not rjson:   # pragma: no cover
            raise SDSSpaceError(message="space refresh error, len of array")

        rjson = rjson[0]

        for label in ['site_is_template',
                      'site_id',
                      'tree_level',
                      'site_name',
                      'site_description',
                      'parent_site_id',
                      'parent_site_name',
                      'site_class_name',
                      'parent_site_class_name',
                      'row_enabled',
                      'multistatus']:
            if label not in rjson:   # pragma: no cover
                raise SDSError("parameter {} not found in space".format(label))
            self.params[label] = rjson[label]

        self.myid = int(self.params['site_id'])

        if 'site_class_parameters' in rjson:
            self.update_class_params(rjson['site_class_parameters'])

    # -------------------------------------
    # def list_block_networks(self, offset=0, page=25, limit=0, collected=0):
    #     """return the list of blocks"""
    #     params = {
    #         'limit': page,
    #         'offset': offset
    #     }

    #     if limit > 0:
    #         if page > limit:
    #             params['limit'] = limit

    #     params['WHERE'] = "site_id='{}'".format(self.myid)
    #     params['WHERE'] += " and subnet_level='0' and is_terminal='0'"

    #     try:
    #         rjson = self.sds.query("ip_subnet_list",
    #                                params=params)
    #     except SDSEmptyError:
    #         return None

    #     if 'errmsg' in rjson:  # pragma: no cover
    #         raise SDSNetworkError(message="net list, "
    #                               + rjson['errmsg'])

    #     anets = []
    #     for net in rjson:
    #         anets.append({
    #             'start_hostaddr': net['start_hostaddr'],
    #             'subnet_size': 32-int(math.log(int(net['subnet_size']), 2)),
    #             'subnet_name': net['subnet_name']
    #         })

    #     # no limit, we should get all the records
    #     if len(rjson) == page:
    #         if limit == 0 or collected < limit:
    #             newnets = self.list_block_networks(offset+page,
    #                                                page=page,
    #                                                limit=limit,
    #                                                collected=(len(anets)
    #                                                           + collected))
    #             if newnets is not None:
    #                 anets += newnets

    #     if limit and len(anets) > limit:
    #         anets = anets[:limit]

    #     return anets

    # -------------------------------------
    def __str__(self):
        """return the string notation of the space object"""
        return_val = "*space* name={}".format(self.name)

        if self.myid != -1:
            return_val += " id={}".format(self.myid)

        if self.params['parent_site_id'] is not None:
            return_val += " parent={}".format(self.params['parent_site_id'])

        return_val += str(super(Space, self).__str__())

        return return_val



import io
import os

from setuptools import setup

here = os.path.dirname(__file__)

with open(os.path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='SOLIDserverRest',
    version='2.2.1',
    author='Gregory CUGAT / Alex Chauvin',
    url='https://gitlab.com/efficientip/solidserverrest',
    description='The SOLIDserverRest is a library to drive EfficientIP API',
    long_description_content_type="text/markdown",
    long_description=long_description,
    author_email='gregory.cugat@efficientip.com, ach@efficientip.com',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    install_requires=['requests~=2.25',
                      'PySocks~=1.7',
                      'urllib3~=1.26',
                      'idna~=2.10',
                      'chardet~=3.0',
                      'pyopenssl~=19.0',
                      'packaging~=20.9'],
    license='BSD 2',
    packages=['SOLIDserverRest', 'SOLIDserverRest.adv'],
    zip_safe=False,
    python_requires=">=3.6",
    py_modules=['check_python_versions'],
    entry_points={
        'console_scripts': [
            'check-python-versions = check_python_versions:main',
        ], }
)



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-03-04 16:47:20 alex>
#
# only for python v3

"""
SOLIDserver management server access
"""

import logging
import ipaddress
import socket

from json.decoder import JSONDecodeError

from SOLIDserverRest.Exception import SDSInitError, SDSAuthError
from SOLIDserverRest.Exception import SDSEmptyError
from SOLIDserverRest import SOLIDserverRest

from .class_params import ClassParams

__all__ = ["SDS"]


# more than 7 arguments to class
# pylint: disable=R0902
class SDS(ClassParams):
    """ class to get connected to a SDS server """

    # ---------------------------
    def __init__(self, ip_address=None, user=None, pwd=None):
        """init the SDS object:
        """
        super(SDS, self).__init__()

        self.sds_ip = None
        if ip_address is not None:
            self.set_server_ip(ip_address)

        self.user = None
        self.pwd = None
        if user is not None and pwd is not None:
            self.set_credentials(user, pwd)

        self.version = None

        self.auth_method = None
        self.check_certificate = False

        self.sds = None
        self.timeout = 1

        self.proxy_socks = None

    # ---------------------------
    def set_server_ip(self, ip_address):
        """set the SOLIDserver IP address for the connection"""
        try:
            ipaddress.IPv4Address(ip_address)
        except ipaddress.AddressValueError as error:
            raise SDSInitError(message="IPv4 address of server error: {}".
                               format(error))

        self.sds_ip = ip_address

    # ---------------------------
    def set_server_name(self, fqdn):
        """set the SOLIDserver FQDN for the connection"""
        # check that the IP exists
        try:
            _ = socket.gethostbyname(fqdn)
        except socket.gaierror as error:
            raise SDSInitError(message="FQDN of the SDS: {}".
                               format(error))

        self.sds_ip = fqdn

    # ---------------------------
    def set_proxy_socks(self, proxy=None):
        """set the SOLIDserver connection through a socks proxy"""
        if proxy:
            self.proxy_socks = proxy

    # ---------------------------
    def set_credentials(self, user=None, pwd=None):
        """add user and login to credentials of this session"""
        if user is None or pwd is None:
            msg = "missing user or password in credentials"
            raise SDSInitError(message=msg)
        self.user = user
        self.pwd = pwd

    # ---------------------------
    def connect(self, method="basicauth", cert_file_path=None, timeout=1):
        """connects to SOLIDserver and check everything is OK by
           querying the version of the admin node in the member list

           method -- basicauth (default) or native (header based)
           cert_file_path -- disable SSL check if None (default)
                             file with cert if check enabled
        """
        if self.user is None or self.pwd is None:
            msg = "missing user or password in credentials for connect"
            raise SDSInitError(message=msg)
        if self.sds_ip is None:
            raise SDSInitError(message="missing ip for server for connect")

        self.sds = SOLIDserverRest(self.sds_ip)

        if method == "basicauth":
            self.sds.use_basicauth_sds(self.user, self.pwd)
            self.auth_method = "basic auth"
        elif method == "native":
            self.sds.use_native_sds(self.user, self.pwd)
            self.auth_method = "native"

        # certificate & SSL check
        if cert_file_path is not None:
            self.sds.set_certificate_file(cert_file_path)
            self.check_certificate = True
        else:
            self.sds.set_ssl_verify(False)

        if timeout != self.timeout:
            self.timeout = timeout

        if self.proxy_socks:
            self.sds.set_proxy(self.proxy_socks)

        self.version = self.get_version()

        if self.version is None:   # pragma: no cover
            self.version = "ukn"
            # need to check if a simple call to a space api is working

    # ---------------------------
    def disconnect(self):
        """disconnects from the SOLIDserver"""
        self.sds = None
        self.version = None
        self.sds_ip = None
        self.user = None
        self.pwd = None
        self.auth_method = None
        self.check_certificate = False
        self.timeout = 1

    # ---------------------------
    def get_version(self):
        """get software version of the SDS based on the management platform
        returns version as a string
        """

        if self.sds is None:
            raise SDSEmptyError(message="not connected")

        if self.version is not None:
            return self.version

        j = self.query("member_list",
                       params={
                           'WHERE': 'member_is_me=1',
                       },
                       option=False)

        if j is None:   # pragma: no cover
            logging.error("error in getting answer on version")
            return None

        if 'member_is_me' not in j[0]:   # pragma: no cover
            logging.error("error in getting version")
            return None

        self.version = j[0]['member_version']
        return self.version

    # ---------------------------
    def get_load(self):
        """get cpu, mem, io"""
        if self.sds is None:
            raise SDSEmptyError(message="not connected")

        j = self.query("member_list",
                       params={
                           'WHERE': 'member_is_me=1',
                       },
                       option=False)

        if j is None:   # pragma: no cover
            logging.error("error in getting answer on version")
            return None

        return {
            'cpu': float(j[0]['member_snmp_cpuload_percent']),
            'ioload': int(j[0]['member_snmp_ioload']),
            'mem': int(j[0]['member_snmp_memory']),
            'hdd': int(j[0]['member_snmp_hdd']),
        }

    # ---------------------------
    def query(self, method, params='', option=False, timeout=1):
        """execute a query towards the SDS"""

        if self.sds is None:
            raise SDSEmptyError(message="not connected")

        _timeout = self.timeout
        if timeout not in (self.timeout, 1):
            _timeout = timeout

        try:
            answer_req = self.sds.query(method,
                                        params=params,
                                        option=option,
                                        timeout=_timeout)

            if answer_req.status_code == 401:   # pragma: no cover
                raise SDSAuthError(message="authentication error")

            if answer_req.status_code == 204:
                raise SDSEmptyError(message="204 returned")

            try:
                j = answer_req.json()
                return j
            except JSONDecodeError:   # pragma: no cover
                logging.error("no json in return")
                return None

        except SDSAuthError as error:   # pragma: no cover
            raise SDSAuthError("{}".format(error))

    # ---------------------------
    def __str__(self):
        """return the string notation of the server object"""
        connected = "not connected"
        if self.version:
            connected = "connected version={} auth={}".format(self.version,
                                                              self.auth_method)
        proxy = ""
        if self.proxy_socks:
            proxy = " socks5h://{}".format(self.proxy_socks)

        return "sds ip={}{} cred={} {} [{}]".format(self.sds_ip,
                                                    proxy,
                                                    self.user,
                                                    connected,
                                                    self.sds)



# -*- Mode: Python; python-indent-offset: 4 -*-
#
# Time-stamp: <2021-05-12 09:44:19 alex>
#
# pylint: disable=R0801


"""
SOLIDserver ip address manager

"""

import binascii
import ipaddress
# import logging
import re
import socket

from packaging.version import Version, parse
from SOLIDserverRest.Exception import (SDSError, SDSIpAddressError,
                                       SDSIpAddressNotFoundError)

from .class_params import ClassParams
from .space import Space


class IpAddress(ClassParams):
    """ class to manipulate the SOLIDserver ip address """

    # -------------------------------------
    def __init__(self, sds=None,  # pylint: disable=too-many-arguments
                 space=None,
                 ipv4=None,
                 name=None,
                 class_params=None):
        """init an address object:
        - sds: object SOLIDserver, could be set afterwards
        - space: space object for this network
        """

        super(IpAddress, self).__init__(sds, name)

        if name:
            self.set_class_params({'hostname': name})

        self.clean_params()
        self.set_sds(sds)

        if space and not isinstance(space, Space):
            raise SDSIpAddressError("no valid space provided")

        self.space = space
        # self.name = name
        self.mac = None

        if ipv4 is not None:
            self.ipv4 = self.check_ipv4_format(ipv4)
            if self.ipv4 is None:
                raise SDSIpAddressError("bad ipv4 format")

        if class_params is not None:
            self.set_class_params(class_params)

    # -------------------------------------
    @classmethod
    def check_ipv4_format(cls, addr):
        """ check the ip v4 format """
        ipv4 = None
        try:
            ipv4 = str(ipaddress.IPv4Address(addr))
        except ValueError:
            ipv4 = None

        return ipv4

    # -------------------------------------
    def clean_params(self):
        """ clean the object params """
        super(IpAddress, self).clean_params()

        self.ipv4 = None
        self.space = None
        self.mac = None

    # -------------------------------------
    def create(self):
        """ create the ip address in SDS """

        if self.sds is None:
            raise SDSIpAddressError(message="not connected")

        if self.space is None:
            raise SDSIpAddressError("no space attached to address for create")

        if self.ipv4 is None:
            raise SDSIpAddressError("no address for create")

        # if object already created
        if self.myid > 0:
            return

        params = {
            'hostaddr': self.ipv4,
            'site_id': self.space.params['site_id'],
            **self.additional_params
        }

        if self.name is not None:
            params['name'] = self.name

        if self.mac is not None:
            params['mac_addr'] = self.mac

        self.prepare_class_params('ip', params)

        # logging.info(params)

        rjson = self.sds.query("ip_address_create",
                               params=params)

        if 'errmsg' in rjson:
            raise SDSIpAddressError(message="creation, "
                                    + rjson['errmsg'])

        self.params['ip_id'] = int(rjson[0]['ret_oid'])
        self.myid = int(self.params['ip_id'])

        self.refresh()

    # -------------------------------------
    def get_id_by_ipaddr(self, ipaddr):
        """get the ID from its ip addr, return None if non existant"""

        params = {
            "limit": 1,
            **self.additional_params
        }
        if parse(self.sds.get_version()) >= Version("7.0.0"):
            params.update({"WHERE": "hostaddr='{}'".format(ipaddr)})
        else:  # pragma: no cover
            params.update({"WHERE": "ip_addr='{}'".format(
                binascii.hexlify(socket.inet_aton(ipaddr)).decode('ascii'))})

        try:
            rjson = self.sds.query('ip_address_list',
                                   params=params)
        except SDSError as err_descr:
            msg = "cannot found object by ip addr {}".format(ipaddr)
            msg += " / "+str(err_descr)
            raise SDSIpAddressNotFoundError(msg)

        if rjson[0]['errno'] != '0':  # pragma: no cover
            raise SDSError("errno raised on get id by addr")

        return rjson[0]['ip_id']

    # -------------------------------------
    def refresh(self):
        """refresh content of the ip address from the SDS"""

        if self.sds is None:
            raise SDSIpAddressError(message="not connected")

        try:
            ip_id = self.get_id_by_ipaddr(ipaddr=self.ipv4)
        except SDSError as err_descr:
            msg = "cannot get ip addr id"
            msg += " / "+str(err_descr)
            raise SDSIpAddressNotFoundError(msg)

        params = {
            "ip_id": ip_id,
            **self.additional_params
        }

        rjson = self.sds.query("ip_address_info",
                               params=params)

        rjson = rjson[0]
        # logging.info(rjson)

        labels = ['ip_id',
                  'name',
                  'mac_addr',
                  'subnet_id',
                  'subnet_size',
                  'subnet_is_terminal',
                  'subnet_start_ip_addr',
                  'subnet_end_ip_addr']
        if parse(self.sds.get_version()) >= Version("7.0.0"):
            labels.extend(['parent_subnet_start_hostaddr',
                           'parent_subnet_end_hostaddr'])

        for label in labels:
            if label not in rjson:  # pragma: no cover
                msg = "parameter {} not found in ip address".format(label)
                raise SDSIpAddressError(msg)
            self.params[label] = rjson[label]

        # logging.warning('update subnet ip in main')
        if 'mac_addr' in self.params:
            if self.params['mac_addr'].startswith('EIP:'):
                del self.params['mac_addr']
            else:
                self.mac = self.params['mac_addr'].lower()

        self.myid = int(rjson['ip_id'])

        if 'ip_class_parameters' in rjson:   # pragma: no cover
            self.update_class_params(rjson['ip_class_parameters'])

    # -------------------------------------
    def delete(self):
        """deletes the ip address in the SDS"""
        if self.sds is None:
            raise SDSIpAddressError(message="not connected")

        if self.myid == -1:
            raise SDSIpAddressNotFoundError("on delete")

        params = {
            'ip_id': self.myid,
            **self.additional_params
        }

        self.sds.query("ip_address_delete",
                       params=params)

        self.clean_params()

    # -------------------------------------
    def update(self):
        """ update the ip address in SDS """

        if self.sds is None:
            raise SDSIpAddressError(message="not connected")

        params = {
            'ip_id': self.myid,
            **self.additional_params
        }

        if self.mac is not None:
            params['mac_addr'] = self.mac

        self.prepare_class_params('ip', params)

        # logging.info(params)

        rjson = self.sds.query("ip_address_update",
                               params=params)

        if 'errmsg' in rjson:  # pragma: no cover
            raise SDSIpAddressError(message="ip addr update error, "
                                    + rjson['errmsg'])

        self.refresh()

    # -------------------------------------
    def set_param(self, param=None, value=None, exclude=None, name=None):
        """ set a specific param on the ip address object """
        super(IpAddress, self).set_param(param,
                                         value,
                                         exclude=['ip_id'],
                                         name='name')

    # -------------------------------------
    def set_name(self, name=None):
        """set the name of the ip address"""
        super(IpAddress, self).set_name(name)
        self.set_class_params({'hostname': name})

    # -------------------------------------
    def set_ipv4(self, addr):
        """set the ip v4 address"""
        self.ipv4 = self.check_ipv4_format(addr)
        if self.ipv4 is None:
            raise SDSIpAddressError("bad ip v4 address format")

    # -------------------------------------
    def set_mac(self, mac):
        """set the mac address linked to this ip"""
        mac = mac.lower()
        if re.match("^[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$",
                    mac):
            mac = re.sub(r'-', ':', mac)
            mac = re.sub(r':', '', mac)
            mac = re.sub(r'(..)(..)(..)(..)(..)(..)',
                         '\\1:\\2:\\3:\\4:\\5:\\6',
                         mac)
            self.mac = mac
            # logging.info(self.mac)
        else:
            self.mac = None
            raise SDSIpAddressError("bad mac format")

    # -------------------------------------
    def __str__(self):  # pragma: no cover
        """return the string notation of the ip address object"""

        return_val = "*ip address*"

        if self.ipv4 is not None:
            return_val += " {}".format(self.ipv4)

        if self.name is not None:
            return_val += " {}".format(self.name)

        if self.mac is not None:
            return_val += " {}".format(self.mac)

        return_val += self.str_params(exclude=['ip_id',
                                               'name'])

        return_val += str(super(IpAddress, self).__str__())

        return return_val
